"""Defines a pass to add a small separation between adjacent joints."""

import argparse
import logging
import xml.etree.ElementTree as ET
from pathlib import Path

import numpy as np
from scipy.spatial.transform import Rotation as R

from onshape.formats.common import save_xml
from onshape.passes.utils import string_to_nparray

logger = logging.getLogger(__name__)


def add_joint_separation(urdf_path: Path, joint_separation_distance: float) -> None:
    """Adds a small separation between joints in a URDF.

    For each joint, moves the joint origin along its local z-axis by the specified distance.
    This helps prevent numerical issues in physics simulation.

    The separation is applied in the joint's local coordinate frame, so it takes into account
    the joint's RPY rotation before applying the translation.

    Args:
        urdf_path: The path to the URDF file.
        joint_separation_distance: Distance in meters to separate joints by.
    """
    urdf_tree = ET.parse(urdf_path)
    root = urdf_tree.getroot()

    for joint in root.findall(".//joint"):
        # Get or create the origin element
        origin = joint.find("origin")
        if origin is None:
            origin = ET.SubElement(joint, "origin")
            origin.attrib["xyz"] = "0 0 0"
            origin.attrib["rpy"] = "0 0 0"

        # Get the joint axis, defaulting to z-axis if not specified
        axis_elem = joint.find("axis")
        if axis_elem is not None:
            axis = string_to_nparray(axis_elem.attrib.get("xyz", "0 0 1"))
            # Normalize the axis vector
            axis = axis / np.linalg.norm(axis)
        else:
            axis = np.array([0, 0, 1])

        # Get current xyz and rpy
        xyz = string_to_nparray(origin.attrib.get("xyz", "0 0 0"))
        rpy = string_to_nparray(origin.attrib.get("rpy", "0 0 0"))

        # Create rotation matrix from rpy
        rotation = R.from_euler("xyz", rpy)

        # Transform the axis vector by the rotation
        transformed_axis = rotation.apply(axis)

        # Scale to desired length and add to current position
        separation_vector = transformed_axis * -joint_separation_distance
        new_xyz = xyz + separation_vector

        # Update the origin xyz
        origin.attrib["xyz"] = " ".join(str(x) for x in new_xyz)

    # Save the modified URDF
    save_xml(urdf_path, urdf_tree)
    logger.info(
        "Added %.3f meter separation to all joints in %s",
        joint_separation_distance,
        urdf_path.name,
    )


def main() -> None:
    parser = argparse.ArgumentParser(description="Add separation between joints in a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    parser.add_argument(
        "--separation",
        type=float,
        default=0.001,
        help="Distance in meters to separate joints by",
    )
    args = parser.parse_args()

    add_joint_separation(args.urdf_path, args.separation)


if __name__ == "__main__":
    # python -m onshape.passes.add_joint_separation
    main()



"""Uses collision meshes as visual meshes in the URDF."""

import argparse
import logging
from pathlib import Path

from onshape.passes.utils import iter_meshes

logger = logging.getLogger(__name__)


def use_collision_meshes_as_visual_meshes(urdf_path: Path) -> None:
    """Updates specified links to use their collision meshes as visual meshes.

    If a link has no collision mesh, its visual mesh will be removed.

    Args:
        urdf_path: The path to the URDF file.
    """
    for link, (visual_mesh, visual_mesh_path), (col_mesh, col_mesh_path) in iter_meshes(
        urdf_path,
        save_when_done=True,
    ):
        name = link.attrib["name"]
        if col_mesh_path is None:
            logger.warning("No collision mesh found for %s, removing visual mesh", name)
            if visual_mesh is not None:
                visual_parent = link.find("visual")
                if visual_parent is not None:
                    link.remove(visual_parent)
            continue

        if visual_mesh is None:
            logger.warning("No visual mesh found for %s", name)
            continue

        if col_mesh is None:
            raise ValueError(f"Missing collision mesh element for {name}")

        if visual_mesh_path != col_mesh_path:
            visual_mesh.attrib["filename"] = col_mesh.attrib["filename"]


def main() -> None:
    parser = argparse.ArgumentParser(description="Use collision meshes as visual meshes in a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    args = parser.parse_args()

    use_collision_meshes_as_visual_meshes(args.urdf_path)


if __name__ == "__main__":
    main()



# mypy: disable-error-code="import-not-found"
"""Simple script to interact with a URDF in PyBullet."""

import argparse
import logging
import math
import time
from pathlib import Path
from typing import Dict, Sequence

from onshape.utils.logging import configure_logging

logger = logging.getLogger(__name__)


def pybullet_main(args: Sequence[str] | None = None) -> None:
    configure_logging()

    parser = argparse.ArgumentParser(description="Show a URDF in PyBullet")
    parser.add_argument("urdf", nargs="?", help="Path to the URDF file")
    parser.add_argument("--dt", type=float, default=0.01, help="Time step")
    parser.add_argument("-n", "--hide-gui", action="store_true", help="Hide the GUI")
    parser.add_argument("--merge-fixed", action="store_true", help="Merge fixed links")
    parser.add_argument("--hide-origin", action="store_true", help="Do not show the origin")
    parser.add_argument("--show-inertia", action="store_true", help="Visualizes the inertia frames")
    parser.add_argument("--see-thru", action="store_true", help="Use see-through mode")
    parser.add_argument("--fixed-base", action="store_true", help="Fix the base linkage in place")
    parser.add_argument("--cycle-duration", type=float, default=10, help="Duration of the joint cycling")
    parser.add_argument("--start-height", type=float, default=1.0, help="Initial height of the robot")
    parsed_args = parser.parse_args(args)

    try:
        import pybullet as p
    except ImportError:
        raise ImportError("pybullet is required to run this script; install it with `pip install pybullet`")

    # Connect to PyBullet.
    p.connect(p.GUI)
    p.setGravity(0, 0, -9.81)
    p.setRealTimeSimulation(0)

    # Turn off panels.
    if parsed_args.hide_gui:
        p.configureDebugVisualizer(p.COV_ENABLE_GUI, 0)
    p.configureDebugVisualizer(p.COV_ENABLE_SEGMENTATION_MARK_PREVIEW, 0)
    p.configureDebugVisualizer(p.COV_ENABLE_DEPTH_BUFFER_PREVIEW, 0)
    p.configureDebugVisualizer(p.COV_ENABLE_RGB_BUFFER_PREVIEW, 0)

    # Enable mouse picking.
    p.configureDebugVisualizer(p.COV_ENABLE_MOUSE_PICKING, 1)

    # Loads the floor plane.
    floor = p.loadURDF(str((Path(__file__).parent / "bullet" / "plane.urdf").resolve()))

    urdf_path = Path("robot" if parsed_args.urdf is None else parsed_args.urdf)
    if urdf_path.is_dir():
        try:
            urdf_path = next(urdf_path.glob("*.urdf"))
        except StopIteration:
            raise FileNotFoundError(f"No URDF files found in {urdf_path}")
    elif not urdf_path.is_file():
        raise FileNotFoundError(urdf_path)

    # Load the robot URDF.
    start_position = [0.0, 0.0, parsed_args.start_height]  # Use the start_height argument
    start_orientation = p.getQuaternionFromEuler([0.0, 0.0, 0.0])
    flags = p.URDF_USE_INERTIA_FROM_FILE
    flags |= p.URDF_USE_SELF_COLLISION_EXCLUDE_PARENT
    if parsed_args.merge_fixed:
        flags |= p.URDF_MERGE_FIXED_LINKS

    robot = p.loadURDF(
        str(urdf_path.resolve().absolute()),
        start_position,
        start_orientation,
        flags=flags,
        useFixedBase=parsed_args.fixed_base,
    )

    # Initializes physics parameters.
    p.changeDynamics(floor, -1, lateralFriction=1, spinningFriction=-1, rollingFriction=-1)
    p.setPhysicsEngineParameter(fixedTimeStep=parsed_args.dt, maxNumCmdPer1ms=1000)

    # Shows the origin of the robot.
    if not parsed_args.hide_origin:
        p.addUserDebugLine([0, 0, 0], [0.1, 0, 0], [1, 0, 0], parentObjectUniqueId=robot, parentLinkIndex=-1)
        p.addUserDebugLine([0, 0, 0], [0, 0.1, 0], [0, 1, 0], parentObjectUniqueId=robot, parentLinkIndex=-1)
        p.addUserDebugLine([0, 0, 0], [0, 0, 0.1], [0, 0, 1], parentObjectUniqueId=robot, parentLinkIndex=-1)

    # Make the robot see-through.
    joint_ids = [i for i in range(p.getNumJoints(robot))] + [-1]
    if parsed_args.see_thru:
        shape_data = p.getVisualShapeData(robot)
        for i in joint_ids:
            prev_color = shape_data[i][-1]
            p.changeVisualShape(robot, i, rgbaColor=prev_color[:3] + (0.9,))

    def draw_box(pt: list[list[float]], color: tuple[float, float, float], obj_id: int, link_id: int) -> None:
        assert len(pt) == 8
        assert all(len(p) == 3 for p in pt)

        p.addUserDebugLine(pt[0], pt[1], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[1], pt[3], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[3], pt[2], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[2], pt[0], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)

        p.addUserDebugLine(pt[0], pt[4], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[1], pt[5], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[2], pt[6], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[3], pt[7], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)

        p.addUserDebugLine(pt[4 + 0], pt[4 + 1], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[4 + 1], pt[4 + 3], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[4 + 3], pt[4 + 2], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)
        p.addUserDebugLine(pt[4 + 2], pt[4 + 0], color, 1, parentObjectUniqueId=obj_id, parentLinkIndex=link_id)

    # Shows bounding boxes around each part of the robot representing the inertia frame.
    if parsed_args.show_inertia:
        for i in joint_ids:
            dynamics_info = p.getDynamicsInfo(robot, i)
            mass = dynamics_info[0]
            if mass <= 0:
                continue
            inertia = dynamics_info[2]

            # Calculate box dimensions.
            ixx, iyy, izz = inertia[0], inertia[1], inertia[2]
            box_scale_x = math.sqrt(6 * (iyy + izz - ixx) / mass) / 2
            box_scale_y = math.sqrt(6 * (ixx + izz - iyy) / mass) / 2
            box_scale_z = math.sqrt(6 * (ixx + iyy - izz) / mass) / 2
            half_extents = [box_scale_x, box_scale_y, box_scale_z]

            # Create box vertices in local inertia frame
            pt = [
                [half_extents[0], half_extents[1], half_extents[2]],
                [-half_extents[0], half_extents[1], half_extents[2]],
                [half_extents[0], -half_extents[1], half_extents[2]],
                [-half_extents[0], -half_extents[1], half_extents[2]],
                [half_extents[0], half_extents[1], -half_extents[2]],
                [-half_extents[0], half_extents[1], -half_extents[2]],
                [half_extents[0], -half_extents[1], -half_extents[2]],
                [-half_extents[0], -half_extents[1], -half_extents[2]],
            ]

            draw_box(pt, (1, 0, 0), robot, i)

    # Show joint controller.
    joints: dict[str, int] = {}
    controls: dict[str, float] = {}
    for i in range(p.getNumJoints(robot)):
        joint_info = p.getJointInfo(robot, i)
        name = joint_info[1].decode("utf-8")
        joint_type = joint_info[2]
        joints[name] = i
        if joint_type == p.JOINT_PRISMATIC:
            joint_min, joint_max = joint_info[8:10]
            controls[name] = p.addUserDebugParameter(name, joint_min, joint_max, 0.0)
        elif joint_type == p.JOINT_REVOLUTE:
            joint_min, joint_max = joint_info[8:10]
            controls[name] = p.addUserDebugParameter(name, joint_min, joint_max, 0.0)

    def reset_joints_to_zero(robot: int, joints: Dict[str, int]) -> None:
        for joint_id in joints.values():
            joint_info = p.getJointInfo(robot, joint_id)
            joint_min, joint_max = joint_info[8:10]
            zero_position = (joint_min + joint_max) / 2
            p.setJointMotorControl2(robot, joint_id, p.POSITION_CONTROL, zero_position)

    def reset_camera(position: int) -> None:
        height = parsed_args.start_height if parsed_args.fixed_base else 0
        camera_positions = {
            1: (2.0, 0, -30, [0, 0, height]),  # Default view
            2: (2.0, 90, -30, [0, 0, height]),  # Side view
            3: (2.0, 180, -30, [0, 0, height]),  # Back view
            4: (2.0, 270, -30, [0, 0, height]),  # Other side view
            5: (2.0, 0, 0, [0, 0, height]),  # Front level view
            6: (2.0, 0, -80, [0, 0, height]),  # Top-down view
            7: (1.5, 45, -45, [0, 0, height]),  # Closer angled view
            8: (3.0, 30, -30, [0, 0, height]),  # Further angled view
            9: (2.0, 0, 30, [0, 0, height]),  # Low angle view
        }

        if position in camera_positions:
            distance, yaw, pitch, target = camera_positions[position]
            p.resetDebugVisualizerCamera(
                cameraDistance=distance,
                cameraYaw=yaw,
                cameraPitch=pitch,
                cameraTargetPosition=target,
            )

    # Run the simulation until the user closes the window.
    last_time = time.time()
    prev_control_values = {k: 0.0 for k in controls}
    cycle_joints = False
    cycle_start_time = 0.0
    cycle_duration = parsed_args.cycle_duration

    while p.isConnected():
        # Reset the simulation if "r" was pressed.
        keys = p.getKeyboardEvents()
        if ord("r") in keys and keys[ord("r")] & p.KEY_WAS_TRIGGERED:
            p.resetBasePositionAndOrientation(robot, start_position, start_orientation)
            p.setJointMotorControlArray(
                robot,
                range(p.getNumJoints(robot)),
                p.POSITION_CONTROL,
                targetPositions=[0] * p.getNumJoints(robot),
            )

        # Reset joints to zero position if "z" was pressed
        if ord("z") in keys and keys[ord("z")] & p.KEY_WAS_TRIGGERED:
            reset_joints_to_zero(robot, joints)
            cycle_joints = False  # Stop joint cycling if it was active

        # Reset camera if number keys 1-9 are pressed
        for i in range(1, 10):
            if ord(str(i)) in keys and keys[ord(str(i))] & p.KEY_WAS_TRIGGERED:
                reset_camera(i)

        # Start/stop joint cycling if "c" was pressed
        if ord("c") in keys and keys[ord("c")] & p.KEY_WAS_TRIGGERED:
            cycle_joints = not cycle_joints
            if cycle_joints:
                cycle_start_time = time.time()
            else:
                # When stopping joint cycling, set joints to their current positions
                for k, v in controls.items():
                    current_position = p.getJointState(robot, joints[k])[0]
                    p.setJointMotorControl2(robot, joints[k], p.POSITION_CONTROL, current_position)

        # Set joint positions.
        if cycle_joints:
            elapsed_time = time.time() - cycle_start_time
            cycle_progress = (elapsed_time % cycle_duration) / cycle_duration
            for k, v in controls.items():
                joint_info = p.getJointInfo(robot, joints[k])
                joint_min, joint_max = joint_info[8:10]
                target_position = joint_min + (joint_max - joint_min) * math.sin(cycle_progress * math.pi)
                p.setJointMotorControl2(robot, joints[k], p.POSITION_CONTROL, target_position)
        else:
            for k, v in controls.items():
                try:
                    target_position = p.readUserDebugParameter(v)
                    if target_position != prev_control_values[k]:
                        prev_control_values[k] = target_position
                        p.setJointMotorControl2(robot, joints[k], p.POSITION_CONTROL, target_position)
                except p.error:
                    logger.debug("Failed to set joint %s", k)
                    pass

        # Step simulation.
        p.stepSimulation()
        cur_time = time.time()
        time.sleep(max(0, parsed_args.dt - (cur_time - last_time)))
        last_time = cur_time


if __name__ == "__main__":
    # python -m onshape.scripts.pybullet
    pybullet_main()



"""OnShape API and client."""

import asyncio
import logging
from typing import Any, BinaryIO, Literal, Mapping

from httpx import AsyncClient

from onshape.onshape.client import DocumentInfo, OnshapeClient, WorkspaceType
from onshape.onshape.schema.assembly import (
    Assembly,
    AssemblyMetadata,
    Part,
    RootAssembly,
    SubAssembly,
)
from onshape.onshape.schema.document import Document
from onshape.onshape.schema.elements import Elements, ElementType
from onshape.onshape.schema.features import Features
from onshape.onshape.schema.part import PartDynamics, PartMetadata, ThumbnailInfo

logger = logging.getLogger(__name__)

RETRY_STATUS_CODES = (400, 429, 500, 502, 503, 504)


def escape_url(s: str) -> str:
    return s.replace("/", "%2f").replace("+", "%2b")


class OnshapeApi:
    def __init__(
        self,
        client: OnshapeClient,
        max_concurrent_requests: int = 1,
        post_wait: float = 0.0,
    ) -> None:
        super().__init__()
        self.client = client
        self.semaphore = asyncio.Semaphore(max_concurrent_requests)
        self.post_wait = post_wait

    def parse_url(self, document_url: str) -> DocumentInfo:
        return self.client.parse_url(document_url)

    async def _request(
        self,
        method: Literal["get", "post", "put", "delete"],
        path: str,
        query: Mapping[str, Any] | None = None,
        headers: Mapping[str, str] | None = None,
        body: Mapping[str, Any] | None = None,
        base_url: str | None = None,
    ) -> dict:
        async with self.semaphore:
            async with self.client.request(
                method=method,
                path=path,
                query={} if query is None else query,
                headers={} if headers is None else headers,
                body={} if body is None else body,
                base_url=base_url,
            ) as response:
                await response.aread()
                response.raise_for_status()
                response_data = response.json()

        if self.post_wait > 0.0:
            await asyncio.sleep(self.post_wait)

        return response_data

    async def get_document(self, did: str) -> Document:
        data = await self._request("get", f"/api/documents/{did}")
        return Document.model_validate(data)

    async def list_elements(
        self,
        document_id: str,
        workspace_id: str,
        workspace_type: WorkspaceType = "w",
    ) -> Elements:
        data = await self._request(
            "get",
            f"/api/documents/d/{document_id}/{workspace_type}/{workspace_id}/elements",
        )
        return Elements.model_validate(data)

    async def get_first_assembly_id(
        self,
        document_id: str,
        workspace_id: str,
        workspace_type: WorkspaceType = "w",
    ) -> str:
        elements = await self.list_elements(document_id, workspace_id, workspace_type)
        for element in elements.root:
            if element.elementType == ElementType.ASSEMBLY:
                logger.info("Found assembly %s", element.name)
                return element.id
        raise ValueError("Assembly not found")

    async def get_assembly(self, document: DocumentInfo, configuration: str = "default") -> Assembly:
        path = (
            f"/api/assemblies/d/{document.document_id}/{document.item_kind}/{document.item_id}/e/{document.element_id}"
        )
        data = await self._request(
            "get",
            path,
            query={
                "includeMateFeatures": "true",
                "includeMateConnectors": "true",
                "includeNonSolids": "true",
                "configuration": configuration,
            },
        )
        return Assembly.model_validate(data)

    async def get_features(self, asm: RootAssembly | SubAssembly) -> Features:
        path = f"/api/assemblies/d/{asm.documentId}/m/{asm.documentMicroversion}/e/{asm.elementId}/features"
        data = await self._request("get", path)
        return Features.model_validate(data)

    async def get_assembly_metadata(
        self,
        assembly: RootAssembly | SubAssembly,
        configuration: str = "default",
    ) -> AssemblyMetadata:
        path = f"/api/metadata/d/{assembly.documentId}/m/{assembly.documentMicroversion}/e/{assembly.elementId}"
        data = await self._request("get", path, query={"configuration": configuration})
        return AssemblyMetadata.model_validate(data)

    async def get_part_metadata(self, part: Part) -> PartMetadata:
        path = (
            f"/api/metadata/d/{part.documentId}/m/{part.documentMicroversion}"
            f"/e/{part.elementId}/p/{escape_url(part.partId)}"
        )
        data = await self._request("get", path, query={"configuration": part.configuration})
        return PartMetadata.model_validate(data)

    async def get_part_mass_properties(self, part: Part) -> PartDynamics:
        path = (
            f"/api/parts/d/{part.documentId}/m/{part.documentMicroversion}"
            f"/e/{part.elementId}/partid/{escape_url(part.partId)}/massproperties"
        )
        data = await self._request(
            "get",
            path,
            query={
                "configuration": part.configuration,
                "useMassPropertyOverrides": True,
                "elementMicroversionId": part.documentMicroversion,
            },
        )
        return PartDynamics.model_validate(data)

    async def download_stl(
        self,
        part: Part,
        fp: BinaryIO,
        *,
        units: str = "meter",
        min_facet_width: float | None = None,
        max_facet_width: float | None = None,
        retries: int = 3,
    ) -> None:
        path = (
            f"/api/parts/d/{part.documentId}/m/{part.documentMicroversion}"
            f"/e/{part.elementId}/partid/{escape_url(part.partId)}/stl"
        )
        query = {
            "mode": "binary",
            "grouping": True,
            "units": units,
            "configuration": part.configuration,
        }
        if min_facet_width is not None:
            query["minFacetWidth"] = min_facet_width
        if max_facet_width is not None:
            query["maxFacetWidth"] = max_facet_width

        for _ in range(retries):
            async with self.semaphore:
                async with self.client.request(
                    "get",
                    path,
                    query=query,
                    headers={"Accept": "*/*"},
                ) as response:
                    data = await response.aread()
                    if response.status_code in RETRY_STATUS_CODES:
                        logger.warning("Retrying download for %s (status code %d)", part.partId, response.status_code)
                        continue
                    response.raise_for_status()
                    fp.write(data)

            if self.post_wait > 0.0:
                await asyncio.sleep(self.post_wait)
            break
        else:
            raise RuntimeError(
                f"Failed to download STL for part {part.partId} after {retries} retries; "
                "are you connected to the internet, and is the Onshape API accessible?"
            )

    async def list_thumbnails(self, document: DocumentInfo) -> ThumbnailInfo:
        path = f"/api/thumbnails/d/{document.document_id}/{document.item_kind}/{document.item_id}"
        data = await self._request("get", path)
        return ThumbnailInfo.model_validate(data)

    async def download_thumbnail(
        self,
        fp: BinaryIO,
        document: DocumentInfo,
        *,
        width: int = 300,
        height: int = 300,
    ) -> None:
        thumbnail_info = await self.list_thumbnails(document)

        size = f"{width}x{height}"
        for thumbnail in thumbnail_info.sizes:
            if thumbnail.size == size:
                break
        else:
            choices = ", ".join(thumbnail.size for thumbnail in thumbnail_info.sizes)
            raise ValueError(f"Thumbnail size {size} not found! Choices are {choices}")

        async with self.semaphore:
            async with AsyncClient(
                timeout=self.client.timeout,
                follow_redirects=False,
            ) as client:
                async with client.stream(
                    "get",
                    thumbnail.href,
                    headers={"Accept": thumbnail.mediaType},
                    data={},
                ) as response:
                    response.raise_for_status()
                    async for chunk in response.aiter_bytes():
                        fp.write(chunk)

        if self.post_wait > 0.0:
            await asyncio.sleep(self.post_wait)



# ruff: noqa: N815
"""Defines the schema for the Features API."""

import functools
from enum import Enum
from typing import Any, Literal

from pydantic import BaseModel


class FeatureType(str, Enum):
    mate = "mate"
    mateRelation = "mateRelation"
    relation = "replicate"
    empty = ""  # This is required when the assembly includes folders.


class InferenceType(str, Enum):
    CENTROID = "CENTROID"


class ParameterMessage(BaseModel):
    parameterId: str
    hasUserCode: bool
    nodeId: str


Parameter = Any


class MateConnectorMessage(BaseModel):
    isHidden: bool
    implicit: bool
    isAuxiliaryTreeMateConnector: bool
    version: int
    featureType: str
    featureId: str
    name: str
    # Seems to be breaking in a more recent version of Pydantic.
    # parameters: list[Parameter] = Field(discriminator="type")


class MateConnector(BaseModel):
    type: int
    typeName: str
    message: MateConnectorMessage


class SubFeature(BaseModel):
    pass


class SuppressionState(BaseModel):
    type: int


class FeatureMessage(BaseModel):
    version: int
    featureType: FeatureType
    featureId: str
    name: str
    parameters: list[Parameter]
    suppressed: bool
    namespace: str
    subFeatures: list[SubFeature]
    returnAfterSubfeatures: bool
    suppressionState: SuppressionState
    hasUserCode: bool
    nodeId: str
    mateConnectors: list[MateConnector] | None = None

    @functools.cached_property
    def parameter_dict(self) -> dict[str, dict]:
        return {param["message"]["parameterId"]: param for param in self.parameters}


class Feature(BaseModel):
    type: int
    typeName: str
    message: FeatureMessage


class FeatureStatus(str, Enum):
    OK = "OK"
    ERROR = "ERROR"


class FeatureStateMessage(BaseModel):
    featureStatus: FeatureStatus
    inactive: bool


class FeatureStateValue(BaseModel):
    type: Literal[1688]
    typeName: Literal["BTFeatureState"]
    message: FeatureStateMessage


class FeatureState(BaseModel):
    key: str
    value: FeatureStateValue


class Features(BaseModel):
    features: list[Feature]
    featureStates: list[FeatureState]
    isComplete: bool
    serializationVersion: str
    sourceMicroversion: str
    rejectMicroversionSkew: bool
    microversionSkew: bool
    libraryVersion: int



"""Defines functions for merging URDF parts at fixed joints."""

import argparse
import logging
import xml.etree.ElementTree as ET
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, get_args

import numpy as np
from scipy.spatial.transform import Rotation as R

from onshape.formats.common import save_xml
from onshape.passes.utils import string_to_nparray
from onshape.utils.geometry import (
    Dynamics,
    apply_transform,
    combine_dynamics,
    combine_meshes,
    matrix_to_moments,
    moments_to_matrix,
    transform_inertia_tensor,
)
from onshape.utils.mesh import load_file, save_file

logger = logging.getLogger(__name__)

MAX_NAME_LENGTH: int = 64

MeshKind = Literal["visual", "collision"]


@dataclass
class JointOrigin:
    xyz: np.ndarray
    rpy: np.ndarray


def mesh_kinds() -> list[MeshKind]:
    return list(get_args(MeshKind))


def find_not_null(element: ET.Element, key: str) -> ET.Element:
    value = element.find(key)
    if value is None:
        raise ValueError(f"{key} not found in {element.tag}")
    return value


def get_part_mesh_path(element: ET.Element, urdf_dir: Path, kind: MeshKind | None = None) -> Path:
    if kind is None:
        mesh_paths = [get_part_mesh_path(element, urdf_dir, mesh_kind) for mesh_kind in mesh_kinds()]
        if any(mesh_path != mesh_paths[0] for mesh_path in mesh_paths):
            raise ValueError(f"Mesh paths for {element.tag} link do not match")
        return mesh_paths[0]

    mesh_elem = find_not_null(element, kind)
    geometry_elem = find_not_null(mesh_elem, "geometry")
    mesh_elem = find_not_null(geometry_elem, "mesh")
    filename = mesh_elem.attrib["filename"]
    return (urdf_dir / filename).resolve()


def get_new_rpy(
    mass_1: float,
    mass_2: float,
    rpy_1: np.ndarray,
    rpy_2: np.ndarray,
) -> np.ndarray:
    """Gets the new roll, pitch and yaw of the combined parts.

    Args:
        mass_1: The mass of the first part.
        mass_2: The mass of the second part.
        rpy_1: The roll, pitch and yaw of the first part.
        rpy_2: The roll, pitch and yaw of the second part.

    Returns:
        The new roll, pitch and yaw of the combined parts.
    """
    rotation_1 = R.from_euler("xyz", rpy_1).as_matrix()
    rotation_2 = R.from_euler("xyz", rpy_2).as_matrix()
    combined_rotation_matrix = (mass_1 * rotation_1 + mass_2 * rotation_2) / (mass_1 + mass_2)
    u, _, vt = np.linalg.svd(combined_rotation_matrix)
    corrected_rotation_matrix = np.dot(u, vt)
    new_rpy = R.from_matrix(corrected_rotation_matrix).as_euler("xyz")
    return new_rpy


def get_color(element: ET.Element) -> np.ndarray:
    visual_elem = find_not_null(element, "visual")
    material_elem = find_not_null(visual_elem, "material")
    color_elem = find_not_null(material_elem, "color")
    rgba = color_elem.attrib["rgba"].split()
    return np.array([float(x) for x in rgba])


def get_child_to_grandchild_joints(root: ET.Element, link: ET.Element) -> list[ET.Element]:
    link_name = link.attrib["name"]
    grandchild_links = []
    for joint in root.findall(".//joint"):
        parent = find_not_null(joint, "parent")
        if parent.attrib["link"] == link_name:
            grandchild_links.append(joint)
    return grandchild_links


def get_link(root: ET.Element, joint: ET.Element, key: str) -> ET.Element:
    link = find_not_null(joint, key)
    link_name = link.attrib["link"]
    link = root.find(f".//link[@name='{link_name}']")
    if link is None:
        raise ValueError(f"Link {link_name} not found")
    return link


def get_dynamics(inertial_link: ET.Element) -> Dynamics:
    mass_elem = find_not_null(inertial_link, "mass")
    origin_elem = find_not_null(inertial_link, "origin")
    inertia_elem = find_not_null(inertial_link, "inertia")

    # Gets the mass from the link.
    mass = float(mass_elem.attrib["value"])

    # Gets the center of mass from the link.
    xyz = string_to_nparray(origin_elem.attrib["xyz"])
    rpy = string_to_nparray(origin_elem.attrib["rpy"])

    # Gets the inertia matrix from the link.
    ixx = float(inertia_elem.attrib["ixx"])
    iyy = float(inertia_elem.attrib["iyy"])
    izz = float(inertia_elem.attrib["izz"])
    ixy = float(inertia_elem.attrib.get("ixy", "0"))
    ixz = float(inertia_elem.attrib.get("ixz", "0"))
    iyz = float(inertia_elem.attrib.get("iyz", "0"))
    inertia = moments_to_matrix(np.array([ixx, iyy, izz, ixy, ixz, iyz]))

    # Converts to the expected dynamics format.
    com = np.array(xyz)
    rotation_matrix = R.from_euler("xyz", rpy).as_matrix()
    inertia = transform_inertia_tensor(inertia, rotation_matrix)

    return Dynamics(mass, com, inertia)


def update_dynamics(dynamics: Dynamics, relative_transform: np.matrix) -> Dynamics:
    com = apply_transform(dynamics.com[None, :], relative_transform)[0]
    inertia = transform_inertia_tensor(dynamics.inertia, relative_transform[:3, :3])
    return Dynamics(dynamics.mass, com, inertia)


def set_dynamics(inertial_link: ET.Element, dynamics: Dynamics) -> None:
    mass_elem = find_not_null(inertial_link, "mass")
    origin_elem = find_not_null(inertial_link, "origin")
    inertia_elem = find_not_null(inertial_link, "inertia")

    mass_elem.attrib["value"] = str(dynamics.mass)
    origin_elem.attrib["xyz"] = " ".join(str(x) for x in dynamics.com)
    inertia_elem.attrib.update(matrix_to_moments(dynamics.inertia))


def get_origin(joint: ET.Element) -> tuple[ET.Element, JointOrigin]:
    origin_element = joint.find("origin")
    if origin_element is None:
        raise ValueError("Origin element not found in joint")
    xyz = string_to_nparray(origin_element.attrib.get("xyz", "0 0 0"))
    rpy = string_to_nparray(origin_element.attrib.get("rpy", "0 0 0"))
    return origin_element, JointOrigin(xyz, rpy)


def set_origin(origin_element: ET.Element, origin: JointOrigin) -> None:
    origin_element.attrib["xyz"] = " ".join(str(x) for x in origin.xyz)
    origin_element.attrib["rpy"] = " ".join(str(x) for x in origin.rpy)


def origin_and_rpy_to_transform(xyz: np.ndarray, rpy: np.ndarray) -> np.matrix:
    relative_rotation_inner = R.from_euler("xyz", rpy).as_matrix()
    relative_rotation = np.eye(4)
    relative_rotation[:3, :3] = relative_rotation_inner
    relative_translation = np.eye(4)
    relative_translation[:3, 3] = xyz
    relative_transform = np.dot(relative_translation, relative_rotation)
    return relative_transform


def fuse_child_into_parent(root: ET.Element, joint: ET.Element, urdf_dir: Path) -> None:
    parent_link = get_link(root, joint, "parent")
    child_link = get_link(root, joint, "child")
    _, origin = get_origin(joint)

    # Gets the mesh paths for the parent and child.
    parent_mesh_path = get_part_mesh_path(parent_link, urdf_dir)
    child_mesh_path = get_part_mesh_path(child_link, urdf_dir)

    parent_mesh = load_file(parent_mesh_path)
    child_mesh = load_file(child_mesh_path)

    # Gets the transformation matrix.
    relative_transform = origin_and_rpy_to_transform(origin.xyz, origin.rpy)

    # Combines the two meshes into a single mesh.
    combined_mesh = combine_meshes(parent_mesh, child_mesh, relative_transform)

    # Combines the inertia of the parent and child links.
    parent_inertial = find_not_null(parent_link, "inertial")
    child_inertial = find_not_null(child_link, "inertial")
    parent_dynamics = get_dynamics(parent_inertial)
    child_dynamics = get_dynamics(child_inertial)
    child_dynamics = update_dynamics(child_dynamics, relative_transform)
    combined_dynamics = combine_dynamics(parent_dynamics, child_dynamics)
    set_dynamics(parent_inertial, combined_dynamics)

    # Delete the parent and chlid meshes and save the new mesh to the parent
    # mesh location.
    parent_mesh_path.unlink()
    child_mesh_path.unlink()
    save_file(combined_mesh, parent_mesh_path)

    # Removes the child link and joint from the URDF.
    root.remove(joint)
    root.remove(child_link)

    # For each child of the child link, update the parent link to be the parent
    # link of the child link, and update the relative origin.
    grandchild_joints = get_child_to_grandchild_joints(root, child_link)
    for grandchild_joint in grandchild_joints:
        if (grandchild_parent := grandchild_joint.find("parent")) is None:
            raise ValueError("Parent not found in grandchild link")
        grandchild_origin_element, grandchild_origin = get_origin(grandchild_joint)

        grandchild_relative_transform = origin_and_rpy_to_transform(grandchild_origin.xyz, grandchild_origin.rpy)
        new_relative_transform = np.dot(relative_transform, grandchild_relative_transform)

        grandchild_parent.attrib["link"] = parent_link.attrib["name"]
        grandchild_origin.xyz = new_relative_transform[:3, 3]
        grandchild_origin.rpy = R.from_matrix(new_relative_transform[:3, :3]).as_euler("xyz")
        set_origin(grandchild_origin_element, grandchild_origin)


def process_fixed_joints(
    urdf_etree: ET.ElementTree,
    urdf_path: Path,
    ignore_merging_fixed_joints: list[str] | None = None,
) -> ET.ElementTree:
    """Iteratively fuses all fixed joints until none remain.

    This greedily fuses all child links into their parent links at fixed joints,
    removing the fixed joints in the process.

    Args:
        urdf_etree: The URDF element tree.
        urdf_path: The path to the URDF file.
        ignore_merging_fixed_joints: The names of the fixed joints to ignore
            when merging.

    Returns:
        The URDF element tree with all fixed joints fused.
    """
    root = urdf_etree.getroot()

    ignore_set = set([] if ignore_merging_fixed_joints is None else ignore_merging_fixed_joints)
    visited_set: set[str] = set()

    starting_fixed_joint_names = [j.attrib["name"] for j in urdf_etree.findall(".//joint[@type='fixed']")]

    while True:
        joints = [j for j in urdf_etree.findall(".//joint[@type='fixed']") if j.attrib["name"] not in visited_set]
        if not joints:
            break
        for joint in joints:
            if joint.attrib["name"] in ignore_set:
                visited_set.add(joint.attrib["name"])
                continue
            fuse_child_into_parent(root, joint, urdf_path.parent)

    missed_ignored_joints = ignore_set - visited_set
    if missed_ignored_joints:
        raise ValueError(
            f"The following fixed joints were not processed: {missed_ignored_joints}. "
            f"The available fixed joints in the URDF are {sorted(starting_fixed_joint_names)}"
        )

    return urdf_etree


def get_merged_urdf(urdf_path: Path, ignore_merging_fixed_joints: list[str] | None = None) -> None:
    """Merges meshes at each fixed joints to avoid collision issues.

    Args:
        urdf_path: The path to the urdf file.
        ignore_merging_fixed_joints: The names of the fixed joints to ignore
            when merging.

    Returns:
        The path to the merged urdf file.
    """
    urdf_tree = ET.parse(urdf_path)

    # Process the fixed joints
    starting_joint_count = len(urdf_tree.findall(".//joint"))
    merged_urdf = process_fixed_joints(urdf_tree, urdf_path, ignore_merging_fixed_joints)
    ending_joint_count = len(merged_urdf.findall(".//joint"))
    logger.info(
        "Removed %d / %d fixed joints (%.4f%% reduction).",
        starting_joint_count - ending_joint_count,
        starting_joint_count,
        100 * (starting_joint_count - ending_joint_count) / starting_joint_count,
    )

    # Save the merged URDF
    save_xml(urdf_path, merged_urdf)


def main() -> None:
    parser = argparse.ArgumentParser(description="Merge fixed joints in a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    args = parser.parse_args()

    get_merged_urdf(args.urdf_path)


if __name__ == "__main__":
    # python -m onshape.passes.merge_fixed_joints
    main()



"""Defines common types and functions for exporting URDFs."""

import io
import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, Self
from xml.sax.saxutils import escape

import numpy as np
from scipy.spatial.transform import Rotation as R

from onshape.formats.common import save_xml


def xml_escape(unescaped: str) -> str:
    return escape(unescaped, entities={"'": "&apos;", '"': "&quot;"})


def format_number(value: float) -> str:
    # Clip epsilon to avoid scientific notation
    if abs(value) < 1e-6:
        value = 0.0
    return f"{value:.8g}"


@dataclass
class Origin:
    xyz: tuple[float, float, float]
    rpy: tuple[float, float, float]

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        origin = ET.Element("origin") if root is None else ET.SubElement(root, "origin")
        origin.set("xyz", " ".join(format_number(v) for v in self.xyz))
        origin.set("rpy", " ".join(format_number(v) for v in self.rpy))
        return origin

    @classmethod
    def from_matrix(cls, matrix: np.matrix) -> Self:
        x = float(matrix[0, 3])
        y = float(matrix[1, 3])
        z = float(matrix[2, 3])
        roll, pitch, yaw = R.from_matrix(matrix[:3, :3]).as_euler("xyz")
        return cls((x, y, z), (roll, pitch, yaw))

    @classmethod
    def zero_origin(cls) -> Self:
        return cls((0.0, 0.0, 0.0), (0.0, 0.0, 0.0))


@dataclass
class Axis:
    xyz: tuple[float, float, float]

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        axis = ET.Element("axis") if root is None else ET.SubElement(root, "axis")
        axis.set("xyz", " ".join(format_number(v) for v in self.xyz))
        return axis


@dataclass
class JointLimits:
    effort: float  # N*m for revolute, N for prismatic
    velocity: float  # rad/s for revolute, m/s for prismatic
    lower: float  # radians for revolute, meters for prismatic
    upper: float  # radians for revolute, meters for prismatic

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        limit = ET.Element("limit") if root is None else ET.SubElement(root, "limit")
        limit.set("effort", format_number(self.effort))
        limit.set("velocity", format_number(self.velocity))
        limit.set("lower", format_number(self.lower))
        limit.set("upper", format_number(self.upper))
        return limit


@dataclass
class JointMimic:
    joint: str
    multiplier: float = 1.0
    offset: float = 0.0

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        mimic = ET.Element("mimic") if root is None else ET.SubElement(root, "mimic")
        mimic.set("joint", self.joint)
        mimic.set("multiplier", format_number(self.multiplier))
        mimic.set("offset", format_number(self.offset))
        return mimic


@dataclass
class JointDynamics:
    damping: float
    friction: float

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        joint = ET.Element("dynamics") if root is None else ET.SubElement(root, "dynamics")
        joint.set("damping", format_number(self.damping))
        joint.set("friction", format_number(self.friction))
        return joint


@dataclass
class BaseJoint(ABC):
    name: str
    parent: str
    child: str
    origin: Origin

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        joint = ET.Element("joint") if root is None else ET.SubElement(root, "joint")
        joint.set("name", self.name)
        joint.set("type", self.joint_type())
        self.origin.to_xml(joint)
        ET.SubElement(joint, "parent", link=self.parent)
        ET.SubElement(joint, "child", link=self.child)
        return joint

    @abstractmethod
    def joint_type(self) -> str: ...


@dataclass
class RevoluteJoint(BaseJoint):
    limits: JointLimits
    axis: Axis  # The axis of rotation
    dynamics: JointDynamics | None = None  # N*m*s/rad for damping, N*m for friction
    mimic: JointMimic | None = None

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        joint = super().to_xml(root)
        self.limits.to_xml(joint)
        self.axis.to_xml(joint)
        if self.dynamics is not None:
            self.dynamics.to_xml(joint)
        if self.mimic is not None:
            self.mimic.to_xml(joint)
        return joint

    def joint_type(self) -> str:
        return "revolute"


@dataclass
class ContinuousJoint(BaseJoint):
    mimic: JointMimic | None = None

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        joint = super().to_xml(root)
        if self.mimic is not None:
            self.mimic.to_xml(joint)
        return joint

    def joint_type(self) -> str:
        return "continuous"


@dataclass
class PrismaticJoint(BaseJoint):
    limits: JointLimits
    axis: Axis  # The axis of translation
    dynamics: JointDynamics | None = None  # N*s/m for damping, N for friction
    mimic: JointMimic | None = None

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        joint = super().to_xml(root)
        self.limits.to_xml(joint)
        self.axis.to_xml(joint)
        if self.dynamics is not None:
            self.dynamics.to_xml(joint)
        if self.mimic is not None:
            self.mimic.to_xml(joint)
        return joint

    def joint_type(self) -> str:
        return "prismatic"


@dataclass
class FixedJoint(BaseJoint):
    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        joint = super().to_xml(root)
        return joint

    def joint_type(self) -> str:
        return "fixed"


@dataclass
class FloatingJoint(BaseJoint):
    mimic: JointMimic | None = None

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        joint = super().to_xml(root)
        if self.mimic is not None:
            self.mimic.to_xml(joint)
        return joint

    def joint_type(self) -> str:
        return "floating"


@dataclass
class PlanarJoint(BaseJoint):
    limits: JointLimits
    axis: Axis  # The surface normal
    mimic: JointMimic | None = None

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        joint = super().to_xml(root)
        self.limits.to_xml(joint)
        self.axis.to_xml(joint)
        if self.mimic is not None:
            self.mimic.to_xml(joint)
        return joint

    def joint_type(self) -> str:
        return "planar"


@dataclass
class Inertia:
    ixx: float
    iyy: float
    izz: float
    ixy: float
    ixz: float
    iyz: float

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        inertia = ET.Element("inertia") if root is None else ET.SubElement(root, "inertia")
        inertia.set("ixx", format_number(self.ixx))
        inertia.set("iyy", format_number(self.iyy))
        inertia.set("izz", format_number(self.izz))
        inertia.set("ixy", format_number(self.ixy))
        inertia.set("ixz", format_number(self.ixz))
        inertia.set("iyz", format_number(self.iyz))
        return inertia


@dataclass
class InertialLink:
    mass: float
    inertia: Inertia
    origin: Origin
    name: str | None = None

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        inertial = ET.Element("inertial") if root is None else ET.SubElement(root, "inertial")
        ET.SubElement(inertial, "mass", value=format_number(self.mass))
        self.inertia.to_xml(inertial)
        self.origin.to_xml(inertial)
        if self.name is not None:
            inertial.set("name", self.name)
        return inertial


@dataclass
class BaseGeometry(ABC):
    @abstractmethod
    def to_xml(self, root: ET.Element | None = None) -> ET.Element: ...


@dataclass
class BoxGeometry(BaseGeometry):
    size: tuple[float, float, float]

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        geometry = ET.Element("geometry") if root is None else ET.SubElement(root, "geometry")
        ET.SubElement(geometry, "box", size=" ".join(format_number(v) for v in self.size))
        return geometry


@dataclass
class CylinderGeometry(BaseGeometry):
    radius: float
    length: float

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        geometry = ET.Element("geometry") if root is None else ET.SubElement(root, "geometry")
        ET.SubElement(
            geometry,
            "cylinder",
            radius=format_number(self.radius),
            length=format_number(self.length),
        )
        return geometry


@dataclass
class SphereGeometry(BaseGeometry):
    radius: float

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        geometry = ET.Element("geometry") if root is None else ET.SubElement(root, "geometry")
        ET.SubElement(geometry, "sphere", radius=format_number(self.radius))
        return geometry


@dataclass
class MeshGeometry(BaseGeometry):
    filename: str
    name: str | None = None

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        geometry = ET.Element("geometry") if root is None else ET.SubElement(root, "geometry")
        if self.name is not None:
            geometry.set("name", self.name)
        ET.SubElement(geometry, "mesh", filename=self.filename)
        return geometry

    def __post_init__(self) -> None:
        self.filename = xml_escape(self.filename)


@dataclass
class Material:
    name: str
    color: list[float]

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        material = ET.Element("material") if root is None else ET.SubElement(root, "material")
        material.set("name", self.name)
        ET.SubElement(material, "color", rgba=" ".join(format_number(v) for v in self.color))
        return material

    @classmethod
    def from_color(cls, color: Literal["red", "green", "blue", "yellow", "white", "black"]) -> Self:
        color_value = {
            "red": [1.0, 0.0, 0.0],
            "green": [0.0, 1.0, 0.0],
            "blue": [0.0, 0.0, 1.0],
            "yellow": [1.0, 1.0, 0.0],
            "white": [1.0, 1.0, 1.0],
            "black": [0.0, 0.0, 0.0],
        }[color]
        return cls(color, color_value)

    def __post_init__(self) -> None:
        if len(self.color) == 3:
            self.color.append(0.5)
        if len(self.color) != 4:
            raise ValueError(f"Color must have 3 or 4 components, got {len(self.color)}")


@dataclass
class VisualLink:
    origin: Origin
    geometry: BaseGeometry
    material: Material
    name: str | None = None

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        visual = ET.Element("visual") if root is None else ET.SubElement(root, "visual")
        if self.name is not None:
            visual.set("name", self.name)
        self.origin.to_xml(visual)
        self.geometry.to_xml(visual)
        self.material.to_xml(visual)
        return visual


@dataclass
class CollisionLink:
    origin: Origin
    geometry: BaseGeometry
    name: str | None = None

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        collision = ET.Element("collision") if root is None else ET.SubElement(root, "collision")
        if self.name is not None:
            collision.set("name", self.name)
        self.origin.to_xml(collision)
        self.geometry.to_xml(collision)
        return collision


@dataclass
class Link:
    name: str
    visual: VisualLink | None = None
    collision: CollisionLink | None = None
    inertial: InertialLink | None = None

    def to_xml(self, root: ET.Element | None = None) -> ET.Element:
        link = ET.Element("link") if root is None else ET.SubElement(root, "link")
        link.set("name", self.name)
        if self.visual is not None:
            self.visual.to_xml(link)
        if self.collision is not None:
            self.collision.to_xml(link)
        if self.inertial is not None:
            self.inertial.to_xml(link)
        return link


@dataclass
class Robot:
    name: str
    parts: list[Link | BaseJoint]

    def to_xml(self) -> ET.Element:
        robot = ET.Element("robot", name=self.name)
        for part in self.parts:
            part.to_xml(robot)
        return robot

    def save(self, path: str | Path | io.StringIO) -> None:
        tree = ET.ElementTree(self.to_xml())
        save_xml(path, tree)



"""Adds a new base linkage to the URDF."""

import logging
import xml.etree.ElementTree as ET
from pathlib import Path

from onshape.formats.common import save_xml

logger = logging.getLogger(__name__)


def add_base_linkage(
    urdf_path: Path,
    base_xyz: tuple[float, float, float] = (0.0, 0.0, 0.0),
    base_rpy: tuple[float, float, float] = (0.0, 0.0, 0.0),
    link_name: str = "base",
    joint_name: str = "base_joint",
    joint_type: str = "fixed",
) -> None:
    """Adds a new base linkage to the URDF.

    Args:
        urdf_path: The path to the URDF file.
        base_xyz: The XYZ to apply to the base linkage to set the robot's position.
        base_rpy: The RPY to apply to the base linkage to orient the robot.
        link_name: The name of the new link to add.
        joint_name: The name of the joint connecting the new link.
        joint_type: The type of the joint (default is 'fixed').
    """
    # Parse the URDF file
    tree = ET.parse(urdf_path)
    root = tree.getroot()

    # Gets the child link name as the first link in the URDF.
    if (child_link := root.find("link")) is None:
        raise ValueError("No child link found in URDF")
    child_link_name = child_link.attrib["name"]

    # Create a new link element
    new_link = ET.Element("link", name=link_name)
    visual = ET.SubElement(new_link, "visual", name=f"{link_name}_visual")
    geometry = ET.SubElement(visual, "geometry", name=f"{link_name}_geometry")
    ET.SubElement(geometry, "sphere", radius="0.01")
    material = ET.SubElement(visual, "material", name=f"{link_name}_material")
    ET.SubElement(material, "color", rgba="1 0 0 1")
    ET.SubElement(
        visual,
        "origin",
        xyz=" ".join(f"{x:.2f}" for x in base_xyz),
        rpy=" ".join(f"{r:.2f}" for r in base_rpy),
    )

    inertial = ET.SubElement(new_link, "inertial", name=f"{link_name}_inertial")
    ET.SubElement(inertial, "mass", value="0.001")
    ET.SubElement(inertial, "inertia", ixx="0.000001", iyy="0.000001", izz="0.000001", ixy="0", ixz="0", iyz="0")
    ET.SubElement(
        inertial,
        "origin",
        xyz=" ".join(f"{x:.2f}" for x in base_xyz),
        rpy=" ".join(f"{r:.2f}" for r in base_rpy),
    )

    # Create a new joint element
    new_joint = ET.Element("joint", name=joint_name, type=joint_type)
    ET.SubElement(new_joint, "parent", link=link_name)
    ET.SubElement(new_joint, "child", link=child_link_name)
    ET.SubElement(
        new_joint,
        "origin",
        xyz=" ".join(f"{x:.2f}" for x in base_xyz),
        rpy=" ".join(f"{r:.2f}" for r in base_rpy),
    )

    # Insert the new link and joint at the start of the URDF
    root.insert(0, new_link)
    root.insert(1, new_joint)

    # Save the modified URDF
    logger.info("Added new base linkage %s with joint %s to %s", link_name, joint_name, urdf_path)
    save_xml(urdf_path, root)


def main() -> None:
    import argparse

    parser = argparse.ArgumentParser(description="Add a new base linkage to a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    parser.add_argument("link_name", type=str, help="The name of the new link to add.")
    parser.add_argument("joint_name", type=str, help="The name of the joint connecting the new link.")
    parser.add_argument(
        "--joint_type",
        type=str,
        default="fixed",
        help="The type of the joint (default is 'fixed')",
    )
    args = parser.parse_args()

    add_base_linkage(args.urdf_path, args.link_name, args.joint_name, args.joint_type)


if __name__ == "__main__":
    # python -m onshape.passes.add_base_linkage
    main()



# ruff: noqa: N815
"""Defines the schema for the Document API."""

from enum import Enum

from pydantic import BaseModel, RootModel


class ElementType(str, Enum):
    ASSEMBLY = "ASSEMBLY"
    PARTSTUDIO = "PARTSTUDIO"
    DRAWING = "DRAWING"
    BLOB = "BLOB"


class Element(BaseModel):
    name: str
    id: str
    type: str
    elementType: ElementType


class Elements(RootModel):
    root: list[Element]



"""Defines common utility functions used by multiple passes."""

import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Any, Iterator

import numpy as np


def iter_meshes(
    urdf_path: Path,
    save_when_done: bool = False,
) -> Iterator[
    tuple[
        ET.Element,
        tuple[ET.Element, Path] | tuple[None, None],
        tuple[ET.Element, Path] | tuple[None, None],
    ]
]:
    urdf_tree = ET.parse(urdf_path)

    def get_mesh(visual_or_collision: ET.Element | None) -> tuple[ET.Element, Path] | tuple[None, None]:
        if visual_or_collision is None:
            return (None, None)
        if (geometry := visual_or_collision.find("geometry")) is None:
            return (None, None)
        if (mesh := geometry.find("mesh")) is None:
            return (None, None)
        return mesh, (urdf_path.parent / mesh.attrib["filename"]).resolve()

    for link in urdf_tree.iter("link"):
        visual_link = link.find("visual")
        collision_link = link.find("collision")

        visual_mesh = get_mesh(visual_link)
        collision_mesh = get_mesh(collision_link)

        yield link, visual_mesh, collision_mesh

    if save_when_done:
        urdf_tree.write(urdf_path, encoding="utf-8", xml_declaration=True)


def string_to_nparray(string: str | bytes | Any) -> np.ndarray:  # noqa: ANN401
    """Converts a string to a numpy array.

    Args:
        string: The string to convert.

    Returns:
        The numpy array.
    """
    if isinstance(string, bytes):
        string = string.decode("utf-8")
    return np.array([float(item) for item in string.split(" ")])



"""Defines a pass to get an MJCF file from the URDF."""

import argparse
import logging
from pathlib import Path

from onshape.formats.common import ActuatorMetadata, JointMetadata
from onshape.formats.mjcf import ConversionMetadata, convert_to_mjcf_metadata

logger = logging.getLogger(__name__)


def convert_joint_type(urdf_type: str) -> str:
    type_map = {
        "revolute": "hinge",
        "prismatic": "slide",
        "continuous": "hinge",
        "fixed": "fixed",
        "floating": "free",
        "planar": "slide",
    }
    return type_map.get(urdf_type, "hinge")


def parse_xyz(xyz_str: str) -> str:
    return " ".join([str(float(x)) for x in xyz_str.split()])


def convert_urdf_to_mjcf(
    urdf_file: str | Path,
    mjcf_file: str | Path | None = None,
    metadata: ConversionMetadata | None = None,
    joint_metadata: dict[str, JointMetadata] | None = None,
    actuator_metadata: dict[str, ActuatorMetadata] | None = None,
) -> list[Path]:
    """Convert URDF to MJCF format.

    Args:
        urdf_file: Path to input URDF file
        mjcf_file: Optional path for output MJCF file
        metadata: Optional metadata for the conversion
        joint_metadata: Optional joint metadata for the conversion
        actuator_metadata: Optional actuator metadata for the conversion

    Returns:
        Path to the generated MJCF file
    """
    urdf_file = Path(urdf_file)
    if mjcf_file is None:
        mjcf_file = urdf_file.with_suffix(".mjcf")
    else:
        mjcf_file = Path(mjcf_file)

    try:
        from urdf2mjcf.convert import convert_urdf_to_mjcf
        from urdf2mjcf.model import ActuatorMetadata, JointMetadata
    except ImportError as e:
        raise ImportError(
            "Please install the package with `urdf2mjcf` as a dependency, using `pip install 'onshape[mujoco]'`"
        ) from e

    mjcf_joint_metadata = (
        None
        if joint_metadata is None
        else {name: JointMetadata.from_dict(meta.to_dict()) for name, meta in joint_metadata.items()}
    )

    mjcf_actuator_metadata = (
        None
        if actuator_metadata is None
        else {typ: ActuatorMetadata.from_dict(meta.to_dict()) for typ, meta in actuator_metadata.items()}
    )

    convert_urdf_to_mjcf(
        urdf_path=urdf_file,
        mjcf_path=mjcf_file,
        metadata=None if metadata is None else convert_to_mjcf_metadata(metadata),
        joint_metadata=mjcf_joint_metadata,
        actuator_metadata=mjcf_actuator_metadata,
    )

    return [mjcf_file]


def main() -> None:
    parser = argparse.ArgumentParser(description="Merge fixed joints in a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    args = parser.parse_args()

    convert_urdf_to_mjcf(args.urdf_path)


if __name__ == "__main__":
    # python -m onshape.passes.simplify_meshes
    main()



"""Excludes specified collision meshes from the URDF."""

import argparse
import logging
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Collection

from onshape.formats.common import save_xml

logger = logging.getLogger(__name__)


def exclude_collision_meshes(urdf_path: Path, exclude_names: Collection[str]) -> None:
    """Removes collision meshes with specified names from the URDF.

    Args:
        urdf_path: The path to the URDF file.
        exclude_names: List of collision mesh names to exclude.
    """
    if len(exclude_names) == 0:
        return

    exclude_names = set(exclude_names)

    urdf_tree = ET.parse(urdf_path)
    root = urdf_tree.getroot()

    all_names: set[str] = set()
    processed: set[str] = set()
    for link in root.findall(".//link"):
        if (name := link.attrib.get("name")) is None:
            continue
        all_names.add(name)
        if name not in exclude_names:
            continue
        for collision in link.findall("collision"):
            link.remove(collision)
            processed.add(name)
            logger.info("Removed collision mesh from link '%s'", name)

    not_processed = exclude_names - processed
    if not_processed:
        all_possible_names_str = "\n".join(sorted(all_names))
        raise ValueError(
            f"Collision meshes not found in URDF: {not_processed}\n"
            f"Here are the possible names:\n{all_possible_names_str}"
        )

    save_xml(urdf_path, urdf_tree)


def main() -> None:
    parser = argparse.ArgumentParser(description="Exclude collision meshes from a URDF.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    parser.add_argument("--exclude", nargs="+", help="Names of collision meshes to exclude.")
    args = parser.parse_args()

    exclude_collision_meshes(args.urdf_path, args.exclude)


if __name__ == "__main__":
    main()



# mypy: disable-error-code="attr-defined"
"""Defines functions for download Onshape links as URDFs.

Note that this module is intended to be used as a stand-alone script which
outputs a URDF file. We can additionally post-process the URDF files using
various downstream scripts, but we should avoid adding any unnecessary behavior
here.
"""

import asyncio
import hashlib
import io
import itertools
import logging
import re
import sys
from collections import deque
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Coroutine, Deque, Iterator, Literal, Sequence, TypeVar

import networkx as nx
import numpy as np
import stl

from onshape.formats import urdf
from onshape.onshape.api import OnshapeApi
from onshape.onshape.cached_api import CachedOnshapeApi
from onshape.onshape.cacher import Cacher
from onshape.onshape.client import DocumentInfo, OnshapeClient
from onshape.onshape.config import DownloadConfig, Joint, JointLimits, MimicRelation
from onshape.onshape.schema.assembly import (
    Assembly,
    AssemblyInstance,
    AssemblyMetadata,
    Instance,
    Key,
    MateFeature,
    MateGroupFeature,
    MateRelationFeature,
    MateType,
    Part,
    PartInstance,
    RelationType,
    SubAssembly,
)
from onshape.onshape.schema.common import ElementUid
from onshape.onshape.schema.document import Document
from onshape.onshape.schema.features import Feature, FeatureStatus
from onshape.onshape.schema.part import PartDynamics, PartMetadata
from onshape.utils.errors import catch_error
from onshape.utils.geometry import apply_matrix_, inv_tf, transform_inertia_tensor
from onshape.utils.logging import configure_logging
from onshape.utils.mesh import COLLISION_SUFFIX
from onshape.utils.resolvers import ExpressionResolver

logger = logging.getLogger(__name__)

EPSILON = 1e-6

T = TypeVar("T")
Tk = TypeVar("Tk")
Tv = TypeVar("Tv")

Color = tuple[int, int, int, int]


def clean_name(name: str) -> str:
    name = name.replace("<", "(").replace(">", ")")
    # name = re.sub(r"[<>]", "", name)
    name = re.sub(r"\s+", "_", name)
    name = re.sub(r"[/\\]+", "_", name)
    return name


def get_configuration_str(configuration: str) -> str:
    if configuration == "default":
        return ""
    if len(configuration) > 40:
        return hashlib.md5(configuration.encode()).hexdigest()[:16]
    return f"_{configuration}"


async def gather_dict(d: dict[Tk, Coroutine[Any, Any, Tv]]) -> dict[Tk, Tv]:
    dict_items = [(k, v) for k, v in d.items()]
    values = await asyncio.gather(*(v for _, v in dict_items))
    return {k: v for (k, _), v in zip(dict_items, values)}


class FailedCheckError(ValueError):
    """An error that occurs when a check fails.

    Args:
        msg: The error message.
        suggestions: Suggestions for fixing the error.
        orig_errs: Original errors that caused the failure.
        end_msg: Additional information to add to the error message.
    """

    def __init__(
        self,
        msg: str,
        *,
        extra_msgs: Sequence[str] | None = None,
        suggestions: Sequence[str] | None = None,
        orig_errs: Sequence[Exception] | None = None,
        end_msg: str | None = None,
    ) -> None:
        self.original_msg = msg
        self.extra_msgs = extra_msgs or []
        self.suggestions = suggestions or []
        self.orig_errs = orig_errs or []
        self.end_msg = end_msg
        full_msg = f"{msg}"
        if self.extra_msgs:
            full_msg += "\n\n" + "\n".join(self.extra_msgs)
        if self.suggestions:
            full_msg += "\n\nSuggestions:" + "".join(f"\n * {s}" for s in self.suggestions)
        if self.orig_errs:
            for orig_err in self.orig_errs:
                full_msg += f"\n\nOriginal error:\n{orig_err}"
        if self.end_msg is not None:
            full_msg += f"\n\n{self.end_msg}"
        super().__init__(full_msg)


def traverse_assemblies(assembly: Assembly) -> Iterator[tuple[Key, AssemblyInstance, SubAssembly]]:
    subassembly_deque: Deque[tuple[Key, SubAssembly]] = deque()
    visited: set[Key] = set()

    # Mapping from EUID to subassembly.
    euid_to_subassembly = {sub.key: sub for sub in assembly.subAssemblies}

    # Adds the root assembly to the traversal.
    for instance in assembly.rootAssembly.instances:
        if isinstance(instance, AssemblyInstance):
            instance_path: Key = (instance.id,)
            if instance_path in visited:
                continue
            visited.add(instance_path)
            subassembly = euid_to_subassembly[instance.euid]
            yield instance_path, instance, subassembly
            subassembly_deque.append((instance_path, subassembly))

    # Adds all the subassemblies to the traversal, recursively.
    while subassembly_deque:
        path, sub_assembly = subassembly_deque.popleft()
        for instance in sub_assembly.instances:
            if isinstance(instance, AssemblyInstance):
                instance_path = path + (instance.id,)
                if instance_path in visited:
                    continue
                visited.add(instance_path)
                subassembly = euid_to_subassembly[instance.euid]
                yield instance_path, instance, subassembly
                subassembly_deque.append((instance_path, subassembly))


def get_key_to_instance_mapping(assembly: Assembly) -> dict[Key, Instance]:
    key_to_part_instance: dict[Key, Instance] = {}
    for instance in assembly.rootAssembly.instances:
        key_to_part_instance[(instance.id,)] = instance
    for path, assembly_instance, sub_assembly in traverse_assemblies(assembly):
        key_to_part_instance[path] = assembly_instance
        for instance in sub_assembly.instances:
            key_to_part_instance[path + (instance.id,)] = instance
    return key_to_part_instance


def get_key_to_part_instance_mapping(key_to_instance_mapping: dict[Key, Instance]) -> dict[Key, PartInstance]:
    return {k: v for k, v in key_to_instance_mapping.items() if isinstance(v, PartInstance)}


def get_key_to_assembly_instance_mapping(key_to_instance_mapping: dict[Key, Instance]) -> dict[Key, AssemblyInstance]:
    return {k: v for k, v in key_to_instance_mapping.items() if isinstance(v, AssemblyInstance)}


def get_key_to_feature(assembly: Assembly) -> dict[Key, MateRelationFeature | MateFeature | MateGroupFeature]:
    feature_mapping: dict[Key, MateRelationFeature | MateFeature | MateGroupFeature] = {}
    for feature in assembly.rootAssembly.features:
        feature_mapping[(feature.id,)] = feature
    for key, _, sub_assembly in traverse_assemblies(assembly):
        for feature in sub_assembly.features:
            feature_mapping[key + (feature.id,)] = feature
    return feature_mapping


def get_key_to_mate_feature(
    key_to_feature: dict[Key, MateRelationFeature | MateFeature | MateGroupFeature],
) -> dict[Key, MateFeature]:
    return {p: f for p, f in key_to_feature.items() if isinstance(f, MateFeature)}


def get_key_to_mate_relation_feature(
    key_to_feature: dict[Key, MateRelationFeature | MateFeature | MateGroupFeature],
) -> dict[Key, MateRelationFeature]:
    return {p: f for p, f in key_to_feature.items() if isinstance(f, MateRelationFeature)}


def get_key_to_name(assembly: Assembly) -> dict[Key, list[str]]:
    key_name_mapping: dict[Key, list[str]] = {}
    key_name_mapping[()] = []
    for instance in assembly.rootAssembly.instances:
        key_name_mapping[(instance.id,)] = [instance.name]
    for feature in assembly.rootAssembly.features:
        key_name_mapping[(feature.id,)] = [feature.featureData.name]
    for key, assembly_instance, sub_assembly in traverse_assemblies(assembly):
        key_name_mapping[key] = key_name_mapping[key[:-1]] + [assembly_instance.name]
        for instance in sub_assembly.instances:
            key_name_mapping[key + (instance.id,)] = key_name_mapping[key] + [instance.name]
        for feature in sub_assembly.features:
            key_name_mapping[key + (feature.id,)] = key_name_mapping[key] + [feature.featureData.name]
    return key_name_mapping


class KeyNamer:
    def __init__(self, assembly: Assembly, joiner: str = ":", clean: bool = True) -> None:
        self.key_to_name = get_key_to_name(assembly)
        self.joiner = joiner
        self.clean = clean

    def lookup_key(
        self,
        name: str,
        prefix: Literal["link", "joint", None],
        joiner: str | None = None,
        clean: bool | None = None,
    ) -> Key:
        for key in self.key_to_name:
            if self(key, prefix, joiner, clean) == name:
                return key
        raise ValueError(f'Name "{name}" not found in key namer.')

    def __call__(
        self,
        key: Key,
        prefix: Literal["link", "joint", None],
        joiner: str | None = None,
        clean: bool | None = None,
    ) -> str:
        if clean is None:
            clean = self.clean
        if joiner is None:
            joiner = self.joiner
        name_parts = self.key_to_name.get(key, key)
        name = joiner.join(name_parts)
        if clean:
            name = clean_name(name)
        return ("" if prefix is None else f"{prefix}{joiner}") + name


def log_graph(graph: nx.Graph, key_namer: KeyNamer) -> None:
    for edge in graph.edges:
        logger.info(
            'Edge: Part "%s" -> Path "%s" (Feature "%s")',
            key_namer(edge[0], "link"),
            key_namer(edge[1], "link"),
            graph.edges[edge]["name"],
        )


def get_graph(
    assembly: Assembly,
    key_namer: KeyNamer | None = None,
    key_to_part_instance: dict[Key, PartInstance] | None = None,
    key_to_mate_feature: dict[Key, MateFeature] | None = None,
) -> tuple[nx.Graph, set[Key]]:
    """Converts the assembly to an undirected graph of joints and parts.

    This checks that the assembly is connected as a single piece (meaning
    that there are no unconnected parts or joints) and that it is a tree
    (meaning that there are no parallel connections). Parallel connections
    arise when a part has two parents.

    Args:
        assembly: The assembly to convert.
        key_namer: The key namer to use.
        key_to_part_instance: The key to part instance mapping.
        key_to_mate_feature: The key to mate feature mapping

    Returns:
        The graph of the assembly, along with a set of the ignored joints.
    """
    graph = nx.Graph()

    if key_namer is None:
        key_namer = KeyNamer(assembly)
    if key_to_part_instance is None:
        key_to_part_instance = get_key_to_part_instance_mapping(get_key_to_instance_mapping(assembly))
    if key_to_mate_feature is None:
        key_to_feature = get_key_to_feature(assembly)
        key_to_mate_feature = get_key_to_mate_feature(key_to_feature)

    # Adds all the graph nodes.
    for key, instance in key_to_part_instance.items():
        if instance.suppressed:
            continue
        graph.add_node(key)

    def add_edge_safe(node_a: Key, node_b: Key, joint: Key) -> None:
        if node_a not in key_to_part_instance:
            logger.warning(
                'Part instance "%s" not found in key to part instance mapping',
                key_namer(node_a, None, " : ", False),
            )
            return
        if node_b not in key_to_part_instance:
            logger.warning(
                'Part instance "%s" not found in key to part instance mapping',
                key_namer(node_b, None, " : ", False),
            )
            return
        part_instance_a = key_to_part_instance[node_a]
        part_instance_b = key_to_part_instance[node_b]
        if part_instance_a.suppressed or part_instance_b.suppressed:
            return

        # Here, we are considering a graph with no ignored joints
        # to make sure the original graph is connected.
        for node_lhs, node_rhs in ((node_a, node_b), (node_b, node_a)):
            if node_lhs not in graph:
                lhs_name = key_namer(node_lhs, None, " : ", False)
                rhs_name = key_namer(node_rhs, None, " : ", False)
                joint_name = key_namer(joint, None, " : ", False)
                raise ValueError(
                    f'Node "{lhs_name}" (which is connected to "{rhs_name}" via "{joint_name}") '
                    "not found in graph. Do you have a mate between a part and the origin?"
                )

        name = key_namer(joint, "joint")
        graph.add_edge(node_a, node_b, name=name)

    # Add edges between nodes that have a feature connecting them.
    for key, mate_feature in key_to_mate_feature.items():
        if mate_feature.suppressed:
            continue
        for mate_pair in itertools.combinations(mate_feature.featureData.matedEntities, 2):
            add_edge_safe(key[:-1] + mate_pair[0].key, key[:-1] + mate_pair[1].key, key)

    # If there are any unconnected nodes in the graph, raise an error.
    if not nx.is_connected(graph):
        num_components = nx.number_connected_components(graph)
        components: list[str] = []
        for component in nx.connected_components(graph):
            component_list = sorted([key_namer(c, None, " : ", False) for c in component])
            components.append("\n      ".join(component_list))
        components_string = "\n\n".join(f"  {i + 1: <3} {component}" for i, component in enumerate(components))
        raise FailedCheckError(
            "The assembly is not fully connected! URDF export requires a single fully-connected robot, "
            f"but the graph has {num_components} components. Ensure that all components have a mate connecting them. "
            f"Components:\n{components_string}"
        )

    if not nx.is_tree(graph):
        # Detects the parallel connections in the graph.
        first_n_bad: list[tuple[str, str]] = []
        for cycle in nx.cycle_basis(graph):
            for i, _ in enumerate(cycle):
                if i == len(cycle) - 1:
                    break
                first_n_bad.append(
                    (
                        key_namer(cycle[i], None, " : ", False),
                        key_namer(cycle[i + 1], None, " : ", False),
                    )
                )
        first_n_bad_str = "\n".join(f" * {a} -> {b}" for a, b in first_n_bad[:5])
        prefix = "first " if len(first_n_bad) > 5 else ""
        raise FailedCheckError(
            f"The assembly has parallel connections! URDF export requires no parallel connections. The {prefix}"
            f"{min(5, len(first_n_bad))} joints in the parallel connection are:\n{first_n_bad_str}",
            end_msg=(
                "A parallel connection occurs when there is a cycle in the graph of parts and joints. URDF files "
                "expect a tree of parts and joints, so this is not allowed. To fix it, you should remove one the "
                "above joints."
            ),
        )

    return graph


def get_digraph(graph: nx.Graph, override_central_node: Key | None = None) -> tuple[nx.DiGraph, Key]:
    """Converts the undirected graph to a directed graph using a root node.

    Args:
        graph: The undirected graph to convert.
        override_central_node: The central node to use.

    Returns:
        The central node of the graph and the directed graph
    """
    # Gets the most central node as the "root" node.
    central_node: Key
    if override_central_node is not None:
        central_node = override_central_node
    else:
        closeness_centrality = nx.closeness_centrality(graph)
        central_node = max(closeness_centrality, key=closeness_centrality.get)
    return nx.bfs_tree(graph, central_node), central_node


def get_joint_list(
    digraph: nx.DiGraph,
    key_to_mate_feature: dict[Key, MateFeature],
    key_namer: KeyNamer,
) -> list[Joint]:
    # Creates a BFS ordering of the graph.
    bfs_node_ordering = list(nx.topological_sort(digraph))
    node_level = {node: i for i, node in enumerate(bfs_node_ordering)}

    # Creates a topologically-sorted list of joints.
    joint_list: list[Joint] = []
    for joint_key, mate_feature in key_to_mate_feature.items():
        if mate_feature.suppressed:
            continue

        if (num_entities := len(mate_feature.featureData.matedEntities)) != 2:
            logger.warning('Mate feature "%s" has %d entity', key_namer(joint_key, None, " : ", False), num_entities)
            continue

        lhs_entity, rhs_entity = mate_feature.featureData.matedEntities
        lhs_key, rhs_key = joint_key[:-1] + lhs_entity.key, joint_key[:-1] + rhs_entity.key

        if lhs_key not in node_level or rhs_key not in node_level:
            lhs_name = key_namer(lhs_key, None, " : ", False)
            rhs_name = key_namer(rhs_key, None, " : ", False)
            logger.warning(
                'Mate feature "%s" has entities "%s" and "%s" that are not connected',
                mate_feature.featureData.name,
                lhs_name,
                rhs_name,
            )
            continue

        lhs_is_first = node_level[lhs_key] < node_level[rhs_key]

        parent_key, child_key = (lhs_key, rhs_key) if lhs_is_first else (rhs_key, lhs_key)
        parent_entity, child_entity = (lhs_entity, rhs_entity) if lhs_is_first else (rhs_entity, lhs_entity)
        mate_type = mate_feature.featureData.mateType
        joint = Joint(
            parent_key,
            child_key,
            parent_entity,
            child_entity,
            mate_type,
            joint_key,
            lhs_is_first,
        )
        joint_list.append(joint)
    joint_list.sort(key=lambda x: (node_level[x.parent], node_level[x.child]))

    return joint_list


async def get_joint_limits(assembly: Assembly, api: OnshapeApi) -> dict[ElementUid, JointLimits]:
    # Gets the features for the assembly and all subassemblies.
    assembly_features = await gather_dict(
        {
            assembly.rootAssembly.key: api.get_features(assembly.rootAssembly),
            **{sub_assembly.key: api.get_features(sub_assembly) for sub_assembly in assembly.subAssemblies},
        }
    )

    joint_limits: dict[ElementUid, JointLimits] = {}

    def get_feature_value(key: str, feature: Feature) -> str | None:
        if (attrib := feature.message.parameter_dict.get(key)) is None:
            return None
        match attrib["typeName"]:
            case "BTMParameterNullableQuantity":
                return None if attrib["message"]["isNull"] else attrib["message"]["expression"]
            case _:
                return None

    for assembly_key, assembly_feature in assembly_features.items():
        for feature_state in assembly_feature.featureStates:
            if feature_state.value.message.featureStatus != FeatureStatus.OK:
                logger.warning(
                    "Feature %s has status %s",
                    feature_state.key,
                    feature_state.value.message.featureStatus,
                )

        for feature in assembly_feature.features:
            key = ElementUid(
                document_id=assembly_key.document_id,
                document_microversion=assembly_key.document_microversion,
                element_id=assembly_key.element_id,
                part_id=feature.message.featureId,
            )

            if (
                "limitsEnabled" not in feature.message.parameter_dict
                or not feature.message.parameter_dict["limitsEnabled"]["message"]["value"]
            ):
                joint_limits[key] = JointLimits()
                continue

            joint_limits[key] = JointLimits(
                z_min_expression=get_feature_value("limitZMin", feature),
                z_max_expression=get_feature_value("limitZMax", feature),
                axial_z_min_expression=get_feature_value("limitAxialZMin", feature),
                axial_z_max_expression=get_feature_value("limitAxialZMax", feature),
            )

    return joint_limits


def get_part_color(part_metadata: PartMetadata) -> Color | None:
    part_color = part_metadata.property_map.get("Appearance")
    if part_color is None:
        return None
    return (
        part_color["color"]["red"],
        part_color["color"]["green"],
        part_color["color"]["blue"],
        part_color["opacity"],
    )


def get_part_name(part: Part, part_metadata: PartMetadata) -> str:
    return str(part_metadata.property_map.get("Name", part.key.unique_id))


async def check_part(
    part: Part,
    api: OnshapeApi,
    *,
    config: DownloadConfig | None = None,
) -> tuple[PartMetadata, PartDynamics]:
    """Checks the metadata and mass properties of a single part.

    Args:
        part: The part to check.
        api: The Onshape API to use.
        config: The download configuration.

    Returns:
        The part metadata and mass properties.

    Raises:
        FailedCheckError: If the part is invalid.
    """
    if config is None:
        config = DownloadConfig()

    part_metadata, part_mass_properties = await asyncio.gather(
        api.get_part_metadata(part),
        api.get_part_mass_properties(part),
    )
    part_name = get_part_name(part, part_metadata)

    if get_part_color(part_metadata) is None:
        raise FailedCheckError(f'Part "{part_name}" has no color. Check that the part has a color.')

    part_dynamic = part_mass_properties.bodies[part.partId]
    mass = part_dynamic.mass[0]

    if mass <= 0 and config.default_part_mass is None:
        raise FailedCheckError(
            f'Part "{part_name}" has a mass of {mass}',
            suggestions=[
                "All parts should have a positive mass.",
                "To fix this, either assign a material to the part or manually set the part mass",
            ],
        )

    return part_metadata, part_mass_properties


async def get_mate_relations(
    key_to_mate_relation_feature: dict[Key, MateRelationFeature],
    *,
    config: DownloadConfig | None = None,
) -> dict[Key, MimicRelation]:
    """Gets the mimic relations from the mate relations.

    Args:
        key_to_mate_relation_feature: The key to mate relation feature mapping.
        config: The download configuration.

    Returns:
        The mimic relations.
    """
    if config is None:
        config = DownloadConfig()
    relations: dict[Key, MimicRelation] = {}
    if config.disable_mimics:
        return relations

    for path, mate_relation_feature in key_to_mate_relation_feature.items():
        if mate_relation_feature.suppressed:
            continue

        relation_type = mate_relation_feature.featureData.relationType
        match relation_type:
            case RelationType.GEAR:
                parent_key, child_key = mate_relation_feature.keys(path[:-1])
                ratio = mate_relation_feature.featureData.relationRatio
                reverse = mate_relation_feature.featureData.reverseDirection
                relations[child_key] = MimicRelation(
                    parent=parent_key,
                    multiplier=-ratio if reverse else ratio,
                )

            case RelationType.LINEAR:
                parent_key, child_key = mate_relation_feature.keys(path[:-1])
                ratio = mate_relation_feature.featureData.relationRatio
                reverse = mate_relation_feature.featureData.reverseDirection
                relations[child_key] = MimicRelation(
                    parent=parent_key,
                    multiplier=-ratio if reverse else ratio,
                )

            case _:
                raise ValueError(f"Unsupported relation type: {relation_type}")

    return relations


@dataclass
class CheckedDocument:
    document: Document
    assembly: Assembly
    assembly_metadata: AssemblyMetadata
    key_to_part_instance: dict[Key, PartInstance]
    key_to_assembly_instance: dict[Key, AssemblyInstance]
    key_to_feature: dict[Key, MateRelationFeature | MateFeature | MateGroupFeature]
    key_to_mate_feature: dict[Key, MateFeature]
    key_to_euid: dict[Key, ElementUid]
    euid_to_part: dict[ElementUid, Part]
    key_namer: KeyNamer
    part_metadata: dict[ElementUid, PartMetadata]
    part_dynamics: dict[ElementUid, PartDynamics]
    central_node: Key
    digraph: nx.DiGraph
    joints: list[Joint]
    joint_limits: dict[ElementUid, JointLimits]
    mate_relations: dict[Key, MimicRelation]


def has_valid_joint_limits(
    joint: Joint,
    joint_limits: dict[ElementUid, JointLimits],
    key_to_euid: dict[Key, ElementUid],
    *,
    config: DownloadConfig | None = None,
) -> bool:
    if config is None:
        config = DownloadConfig()

    joint_assembly_id, feature_id = key_to_euid[joint.joint_key[:-1]], joint.joint_key[-1]
    joint_info_key = ElementUid(
        document_id=joint_assembly_id.document_id,
        document_microversion=joint_assembly_id.document_microversion,
        element_id=joint_assembly_id.element_id,
        part_id=feature_id,
    )

    match joint.mate_type:
        case MateType.REVOLUTE:
            if config.default_revolute_joint_limits is not None:
                return True
            if (limits := joint_limits.get(joint_info_key)) is None:
                return False
            return limits.axial_z_min_expression is not None and limits.axial_z_max_expression is not None
        case MateType.SLIDER:
            if config.default_prismatic_joint_limits is not None:
                return True
            if (limits := joint_limits.get(joint_info_key)) is None:
                return False
            return limits.z_min_expression is not None and limits.z_max_expression is not None
        case _:
            return True


async def check_document(
    document_info: DocumentInfo,
    api: OnshapeApi,
    *,
    config: DownloadConfig | None = None,
) -> CheckedDocument:
    """Checks that a document is valid.

    This function is basically a "pre-processing" step that runs light-weight
    checks on the document to make sure that it can be converted to a URDF
    successfully. We don't download any meshes or do any heavy processing here,
    just some basic checks to make sure that the document is in a good state.

    Args:
        document_info: The document to check.
        api: The Onshape API to use
        config: The converter configuration.

    Raises:
        FailedCheckError: If the document is invalid.
    """
    if config is None:
        config = DownloadConfig()

    # Checks that the document can be downloaded. We do this first to just make
    # sure that the user isn't trying to retrieve a private document.
    try:
        document = await api.get_document(document_info.document_id)
    except Exception as e:
        raise FailedCheckError(
            f"Failed to get document {document_info.get_url()}",
            suggestions=[
                "Check that the document ID is correct.",
                "Check that the document is not private.",
            ],
            orig_errs=(e,),
        ) from e

    # Checks that the assembly is valid.
    try:
        assembly = await api.get_assembly(document_info)
    except Exception as e:
        raise FailedCheckError(
            f"Failed to get assembly for document {document_info.get_url()}",
            suggestions=[
                "Check that the document is an assembly, not a part studio.",
            ],
            orig_errs=(e,),
        ) from e

    # Checks that the assembly metadata is valid.
    try:
        assembly_metadata = await api.get_assembly_metadata(assembly.rootAssembly)
    except Exception as e:
        raise FailedCheckError(
            f"Failed to get assembly metadata for document {document_info.get_url()}",
            suggestions=[
                "Check that the assembly is not empty.",
            ],
            orig_errs=(e,),
        ) from e

    # Gets some mappings from the assembly to make subsequent lookups easier.
    key_to_instance = get_key_to_instance_mapping(assembly)
    key_to_part_instance = get_key_to_part_instance_mapping(key_to_instance)
    key_to_assembly_instance = get_key_to_assembly_instance_mapping(key_to_instance)
    key_to_feature = get_key_to_feature(assembly)
    key_to_mate_feature = get_key_to_mate_feature(key_to_feature)
    key_to_mate_relation_feature = get_key_to_mate_relation_feature(key_to_feature)
    key_to_euid = {key: assembly.key for key, _, assembly in traverse_assemblies(assembly)}
    key_to_euid[()] = assembly.rootAssembly.key
    euid_to_part = {part.key: part for part in assembly.parts}
    key_namer = KeyNamer(assembly)

    try:
        graph = get_graph(assembly, key_namer, key_to_part_instance, key_to_mate_feature)

    except ValueError as e:
        raise FailedCheckError(
            f"Failed to get graph for document {document_info.get_url()}",
            suggestions=[
                "Check that the assembly is fully connected.",
                "Check that there are no parallel connections.",
                "Check that no parts are connected to the origin.",
            ],
            orig_errs=(e,),
        ) from e

    try:
        override_central_node: Key | None
        if config.override_central_node is not None:
            override_central_node = key_namer.lookup_key(config.override_central_node, None, None, False)
        else:
            override_central_node = None
        digraph, central_node = get_digraph(graph, override_central_node)

    except ValueError as e:
        raise FailedCheckError(
            f"Failed to get digraph for document {document_info.get_url()}",
            suggestions=[
                "Check that the provided central node is in the graph.",
            ],
            orig_errs=(e,),
        ) from e

    # Checks all the parts in the assembly.
    check_part_results = await asyncio.gather(
        *(catch_error(check_part(part, api, config=config)) for part in assembly.parts)
    )
    checked_part_properties, errs = zip(*check_part_results)

    if any(err is not None for err in errs):
        extra_msgs = {suggestion for err in errs if isinstance(err, FailedCheckError) for suggestion in err.suggestions}
        raise FailedCheckError(
            f"Invalid parts for document {document_info.get_url()}",
            extra_msgs=[
                *(
                    f"{err.original_msg}" if isinstance(err, FailedCheckError) else f"{type(err).__name__}: {err}"
                    for err in errs
                    if err is not None
                ),
                *extra_msgs,
            ],
            suggestions=[
                "Note that the Onshape API returns a mass of 0 for standard parts.",
                "Check that the part is not a standard part.",
            ],
        )

    part_metadata = {part.key: md for part, (md, _) in zip(assembly.parts, checked_part_properties)}
    part_dynamics = {part.key: dyn for part, (_, dyn) in zip(assembly.parts, checked_part_properties)}

    # Checks all the joints in the assembly.
    joints = get_joint_list(digraph, key_to_mate_feature, key_namer)
    joint_limits = await get_joint_limits(assembly, api)

    # Checks that the joint limits are defined for all relevant joints.
    missing_joint_limits = [
        joint
        for joint in joints
        if not has_valid_joint_limits(
            joint,
            joint_limits,
            key_to_euid,
            config=config,
        )
    ]
    if missing_joint_limits:
        missing_joint_names = "".join(
            f"\n * {key_namer(joint.joint_key, None, ' : ', False)}" for joint in missing_joint_limits
        )
        raise FailedCheckError(
            f"Missing joint limits for {len(missing_joint_limits)} joints: {missing_joint_names}",
            suggestions=[
                "Check that the joint limits are defined for all relevant joints.",
                "Replicating joints currently seems to be unsupported",
            ],
        )

    # Checks all the mate relations in the assembly.
    try:
        mate_relations = await get_mate_relations(key_to_mate_relation_feature, config=config)

    except Exception as e:
        raise FailedCheckError(
            f"Failed to get mate relations for document {document_info.get_url()}",
            suggestions=[
                "Check that you are only using supported mimic relations.",
            ],
            orig_errs=(e,),
        ) from e

    return CheckedDocument(
        document=document,
        assembly=assembly,
        assembly_metadata=assembly_metadata,
        key_to_part_instance=key_to_part_instance,
        key_to_assembly_instance=key_to_assembly_instance,
        key_to_feature=key_to_feature,
        key_to_mate_feature=key_to_mate_feature,
        key_to_euid=key_to_euid,
        euid_to_part=euid_to_part,
        part_metadata=part_metadata,
        part_dynamics=part_dynamics,
        key_namer=key_namer,
        central_node=central_node,
        digraph=digraph,
        joints=joints,
        joint_limits=joint_limits,
        mate_relations=mate_relations,
    )


async def download_stl(
    doc: CheckedDocument,
    key: Key,
    part_file_path: Path,
    api: OnshapeApi,
    stl_origin_to_part_tf: np.ndarray,
    min_facet_width: float | None = None,
    retries: int = 3,
) -> Path:
    part_instance = doc.key_to_part_instance[key]
    part = doc.euid_to_part[part_instance.euid]

    buffer = io.BytesIO()
    await api.download_stl(part, buffer, min_facet_width=min_facet_width, retries=retries)
    buffer.seek(0)
    mesh_obj = stl.mesh.Mesh.from_file(None, fh=buffer)
    mesh_obj = apply_matrix_(mesh_obj, stl_origin_to_part_tf)
    mesh_obj.save(part_file_path)

    return part_file_path


def get_urdf_part(
    doc: CheckedDocument,
    key: Key,
    mesh_dir_name: str,
    joint: Joint | None = None,
    *,
    config: DownloadConfig | None = None,
) -> tuple[urdf.Link, np.ndarray, str]:
    """Returns the URDF link for a part.

    Args:
        doc: The checked document.
        key: The part key.
        mesh_dir: The directory to save the mesh files.
        mesh_dir_name: The name of the mesh directory.
        joint: The joint to use.
        config: The converter configuration.

    Returns:
        The URDF link and the transformation matrix from the STL origin to
        the part frame.
    """
    if config is None:
        config = DownloadConfig()

    part_name = doc.key_namer(key, None)
    part_instance = doc.key_to_part_instance[key]
    part = doc.euid_to_part[part_instance.euid]

    part_metadata = doc.part_metadata[part_instance.euid]
    if (part_color := get_part_color(part_metadata)) is None:
        raise ValueError(f"Part {part_name} has no color.")
    part_dynamic = doc.part_dynamics[part_instance.euid].bodies[part.partId]

    # If the part is the root part, move the STL so that its origin is at the
    # center of mass.
    if joint is None:
        link_to_stl_origin_tf = np.eye(4)
        link_to_stl_origin_tf[:3, 3] = np.array(part_dynamic.center_of_mass).reshape(3)
    else:
        link_to_stl_origin_tf = joint.child_entity.matedCS.part_to_mate_tf
    stl_origin_to_link_tf = inv_tf(np.matrix(link_to_stl_origin_tf))

    # Gets the part mass.
    mass = part_dynamic.mass[0]
    uses_default_mass = False
    if mass <= 0.0:
        if config.default_part_mass is None:
            raise ValueError(f"Part {part_name} has a mass of {mass}.")
        logger.warning("Part %s has a mass of %f, using default mass of %f", part_name, mass, config.default_part_mass)
        mass = config.default_part_mass
        uses_default_mass = True

    # We update the STL mesh so that its origin is at the joint frame.
    mesh_origin = urdf.Origin.zero_origin()
    center_of_mass = part_dynamic.center_of_mass_in_frame(stl_origin_to_link_tf)

    if uses_default_mass:
        inertia_transformed = np.eye(3)
    else:
        inertia = part_dynamic.inertia_matrix
        inertia_transformed = transform_inertia_tensor(inertia, np.matrix(stl_origin_to_link_tf[:3, :3]))

    # Ensures that the inertia tensor is positive definite.
    np.fill_diagonal(inertia_transformed, np.maximum(np.diagonal(inertia_transformed), config.min_inertia_value))

    # Since we are already transforming the inertia tensor so that the
    # principal axes are aligned with the link frame, we can just use the
    # identity matrix for the principal axes.
    # principal_axes = part_dynamic.principal_axes_in_frame(stl_origin_to_link_tf)
    # principal_axes_rpy = R.from_matrix(principal_axes).as_euler("xyz", degrees=False)
    principal_axes_rpy = (0.0, 0.0, 0.0)

    # Gets the part name.
    configuration = part_instance.configuration
    part_file_name = f"{part_name}{get_configuration_str(configuration)}.stl"
    urdf_file_path = f"{mesh_dir_name}/{part_file_name}"
    urdf_link_name = doc.key_namer(key, "link")

    urdf_part_link = urdf.Link(
        name=urdf_link_name,
        visual=urdf.VisualLink(
            origin=mesh_origin,
            geometry=urdf.MeshGeometry(
                name=f"{urdf_link_name}_geometry",
                filename=urdf_file_path,
            ),
            material=urdf.Material(
                name=f"{urdf_link_name}_material",
                color=[c / 255.0 for c in part_color],
            ),
            name=f"{urdf_link_name}_visual",
        ),
        inertial=urdf.InertialLink(
            origin=urdf.Origin(
                xyz=center_of_mass,
                rpy=principal_axes_rpy,
            ),
            mass=mass,
            inertia=urdf.Inertia(
                ixx=float(inertia_transformed[0, 0]),
                ixy=float(inertia_transformed[0, 1]),
                ixz=float(inertia_transformed[0, 2]),
                iyy=float(inertia_transformed[1, 1]),
                iyz=float(inertia_transformed[1, 2]),
                izz=float(inertia_transformed[2, 2]),
            ),
            name=f"{urdf_link_name}_inertial",
        ),
        collision=urdf.CollisionLink(
            origin=mesh_origin,
            geometry=urdf.MeshGeometry(
                name=f"{urdf_link_name}_collision_geometry",
                filename=urdf_file_path,
            ),
            name=f"{urdf_link_name}{COLLISION_SUFFIX}",
        ),
    )

    return urdf_part_link, stl_origin_to_link_tf, configuration


def get_urdf_joint(
    doc: CheckedDocument,
    joint: Joint,
    parent_stl_origin_to_part_tf: np.ndarray,
    *,
    config: DownloadConfig | None = None,
) -> urdf.BaseJoint:
    """Returns the URDF joint.

    Args:
        doc: The checked document.
        joint: The joint to convert.
        parent_stl_origin_to_part_tf: The transformation matrix from the parent
            STL origin to the parent part frame.
        config: The converter configuration.

    Returns:
        The URDF link and joint.
    """
    if config is None:
        config = DownloadConfig()

    parent_part_to_mate_tf = joint.parent_entity.matedCS.part_to_mate_tf
    parent_stl_origin_to_mate_tf = parent_stl_origin_to_part_tf @ parent_part_to_mate_tf

    # Gets the joint limits.
    joint_assembly_id, feature_id = doc.key_to_euid[joint.joint_key[:-1]], joint.joint_key[-1]
    joint_info_key = ElementUid(
        document_id=joint_assembly_id.document_id,
        document_microversion=joint_assembly_id.document_microversion,
        element_id=joint_assembly_id.element_id,
        part_id=feature_id,
    )
    joint_limits = doc.joint_limits.get(joint_info_key)
    expression_resolver = ExpressionResolver(joint_assembly_id.configuration)

    def resolve(expression: str | None) -> float | None:
        return None if expression is None else expression_resolver.read_expression(expression)

    def get_effort_and_velocity(name: str, default_effort: float, default_velocity: float) -> tuple[float, float]:
        effort = default_effort
        velocity = default_velocity

        dof_name = name.split(":")[-1]

        if config.joint_metadata is None:
            logger.warning(
                "Missing joint metadata, using default effort: %s & velocity: %s", default_effort, default_velocity
            )
            return default_effort, default_velocity
        if dof_name not in config.joint_metadata:
            raise ValueError(f"Joint {name} not found in joint metadata")

        actuator_type = config.joint_metadata[dof_name].actuator_type
        if actuator_type is None:
            raise ValueError(f"Joint {name} must have an actuator type")
        if actuator_type not in config.actuators_metadata:
            available_actuator_types = ", ".join(config.actuators_metadata.keys())
            raise ValueError(
                f"Actuator type {actuator_type} for joint {name} not found in actuator metadata. "
                f"Available actuator types: {available_actuator_types or 'none'}"
            )

        actuator = config.actuators_metadata[actuator_type]
        if actuator.max_torque <= 0:
            raise ValueError(f"Actuator {actuator_type} for joint {name} must have a positive max torque")
        if actuator.max_velocity <= 0:
            raise ValueError(f"Actuator {actuator_type} for joint {name} must have a positive max velocity")

        effort = actuator.max_torque
        velocity = actuator.max_velocity

        return effort, velocity

    def get_joint_limits(
        joint_limits: JointLimits | None,
        default_limits: tuple[float, float] | None,
        is_axial: bool,
    ) -> tuple[float | None, float | None]:
        min_value: float | None = None
        max_value: float | None = None
        if joint_limits is not None:
            if is_axial:
                min_value = resolve(joint_limits.axial_z_min_expression)
                max_value = resolve(joint_limits.axial_z_max_expression)
            else:
                min_value = resolve(joint_limits.z_min_expression)
                max_value = resolve(joint_limits.z_max_expression)
        if min_value is None or max_value is None:
            new_min_value, new_max_value = (None, None) if default_limits is None else default_limits
            if min_value is None:
                min_value = new_min_value
            if max_value is None:
                max_value = new_max_value
        return min_value, max_value

    name = doc.key_namer(joint.joint_key, "joint")
    origin = urdf.Origin.from_matrix(parent_stl_origin_to_mate_tf)
    mate_type = joint.mate_type

    match mate_type:
        case MateType.FASTENED:
            parent, child = doc.key_namer(joint.parent, "link"), doc.key_namer(joint.child, "link")
            return urdf.FixedJoint(
                name=name,
                parent=parent,
                child=child,
                origin=origin,
            )

        case MateType.REVOLUTE:
            parent, child = doc.key_namer(joint.parent, "link"), doc.key_namer(joint.child, "link")
            mimic_joint = doc.mate_relations.get(joint.joint_key)

            min_value, max_value = get_joint_limits(joint_limits, config.default_revolute_joint_limits, is_axial=True)

            if min_value is None or max_value is None:
                raise ValueError(f"Revolute joint {name} ({parent} -> {child}) does not have limits defined.")
            if min_value >= max_value - EPSILON:
                raise ValueError(f"Revolute joint {name} ({parent} -> {child}) has range [{min_value}, {max_value}].")

            # We tried this approach at one point, but it was better to modify
            # the axis orientation instead, because it makes it easier to
            # offset the joints in the model.
            # if joint.lhs_is_first:
            #     min_value, max_value = -max_value, -min_value

            effort, velocity = get_effort_and_velocity(
                name,
                config.default_revolute_joint_effort,
                config.default_revolute_joint_velocity,
            )

            return urdf.RevoluteJoint(
                name=name,
                parent=parent,
                child=child,
                origin=origin,
                axis=urdf.Axis((0.0, 0.0, -1.0)) if joint.lhs_is_first else urdf.Axis((0.0, 0.0, 1.0)),
                limits=urdf.JointLimits(
                    effort=effort,
                    velocity=velocity,
                    lower=min_value,
                    upper=max_value,
                ),
                mimic=(
                    None
                    if mimic_joint is None
                    else urdf.JointMimic(
                        joint=doc.key_namer(mimic_joint.parent, "joint"),
                        multiplier=mimic_joint.multiplier,
                        offset=0.0,
                    )
                ),
            )

        case MateType.SLIDER:
            parent, child = doc.key_namer(joint.parent, "link"), doc.key_namer(joint.child, "link")
            mimic_joint = doc.mate_relations.get(joint.joint_key)

            min_value, max_value = get_joint_limits(joint_limits, config.default_prismatic_joint_limits, is_axial=False)

            if min_value is None or max_value is None:
                raise ValueError(f"Slider joint {name} ({parent} -> {child}) does not have limits defined.")
            if min_value >= max_value - EPSILON:
                raise ValueError(f"Slider joint {name} ({parent} -> {child}) has range [{min_value}, {max_value}].")

            # if joint.lhs_is_first:
            #     min_value, max_value = -max_value, -min_value

            effort, velocity = get_effort_and_velocity(
                name,
                config.default_prismatic_joint_effort,
                config.default_prismatic_joint_velocity,
            )

            return urdf.PrismaticJoint(
                name=name,
                parent=parent,
                child=child,
                origin=origin,
                axis=urdf.Axis((0.0, 0.0, -1.0)) if joint.lhs_is_first else urdf.Axis((0.0, 0.0, 1.0)),
                limits=urdf.JointLimits(
                    effort=effort,
                    velocity=velocity,
                    lower=min_value,
                    upper=max_value,
                ),
                mimic=(
                    None
                    if mimic_joint is None
                    else urdf.JointMimic(
                        joint=doc.key_namer(mimic_joint.parent, "joint"),
                        multiplier=mimic_joint.multiplier,
                        offset=0.0,
                    )
                ),
            )

        case MateType.PLANAR:
            raise NotImplementedError

        case MateType.CYLINDRICAL:
            raise NotImplementedError

        case MateType.PIN_SLOT:
            raise NotImplementedError

        case MateType.BALL:
            raise NotImplementedError

        case MateType.PARALLEL:
            raise NotImplementedError

        case _:
            raise ValueError(f"Unsupported mate type: {mate_type}")


@dataclass
class SavedUrdf:
    urdf_path: Path
    mesh_paths: list[Path]


async def save_urdf(
    doc: CheckedDocument,
    output_dir: Path,
    api: OnshapeApi,
    *,
    config: DownloadConfig | None = None,
) -> SavedUrdf:
    """Saves the URDF files for the document.

    Args:
        doc: The checked document to save.
        output_dir: The output directory.
        api: The Onshape API to use.
        config: The converter configuration.

    Returns:
        The path to the saved URDF file and mesh files.
    """
    if config is None:
        config = DownloadConfig()
    mesh_dir_name = config.mesh_dir

    # Gets the root link in the URDF.
    urdf_parts: list[urdf.Link | urdf.BaseJoint] = []
    part_link, root_stl_origin_to_part_tf, configuration = get_urdf_part(
        doc=doc,
        key=doc.central_node,
        mesh_dir_name=mesh_dir_name,
    )
    urdf_parts.append(part_link)

    # Keeps track of the parent STL origin to part transformation matrices.
    stl_origin_to_part_tfs: dict[Key, tuple[np.ndarray, str]] = {
        doc.central_node: (root_stl_origin_to_part_tf, configuration)
    }

    for joint in doc.joints:
        joint_tf, _ = stl_origin_to_part_tfs[joint.parent]
        urdf_joint = get_urdf_joint(doc, joint, joint_tf, config=config)
        urdf_link, stl_origin_to_part_tf, configuration = get_urdf_part(
            doc=doc,
            key=joint.child,
            mesh_dir_name=mesh_dir_name,
            joint=joint,
            config=config,
        )
        stl_origin_to_part_tfs[joint.child] = (stl_origin_to_part_tf, configuration)
        urdf_parts.extend([urdf_joint, urdf_link])

    robot_name = clean_name(str(doc.assembly_metadata.property_map.get("Name", "robot"))).lower()
    urdf_path = output_dir / f"{robot_name if config.use_assembly_name else 'robot'}.urdf"
    urdf_robot = urdf.Robot(name=robot_name, parts=urdf_parts)
    urdf_robot.save(urdf_path)

    # Finally, downloads the STL files.
    (mesh_dir := output_dir / mesh_dir_name).mkdir(parents=True, exist_ok=True)
    mesh_paths = await asyncio.gather(
        *(
            download_stl(
                doc,
                key,
                mesh_dir / f"{doc.key_namer(key, None)}{get_configuration_str(configuration)}.stl",
                api,
                stl_origin_to_part_tf,
                min_facet_width=config.min_facet_width,
            )
            for key, (stl_origin_to_part_tf, configuration) in stl_origin_to_part_tfs.items()
        )
    )

    return SavedUrdf(
        urdf_path=urdf_path,
        mesh_paths=mesh_paths,
    )


@dataclass
class DownloadedDocument:
    check_document: CheckedDocument
    urdf_info: SavedUrdf


async def download(
    document_url: str,
    output_dir: str | Path,
    *,
    config: DownloadConfig | None = None,
    api: OnshapeApi | None = None,
) -> DownloadedDocument:
    """Converts an Onshape document to a URDF.

    Args:
        document_url: The URL of the Onshape document.
        output_dir: The output directory.
        config: The converter configuration.
        api: The Onshape API to use.

    Returns:
        The downloaded document information.
    """
    output_dir = Path(output_dir).expanduser().resolve()
    if config is None:
        config = DownloadConfig()

    # Creates directories for storing cached artifacts and meshes.
    (cache_dir := output_dir / ".cache").mkdir(parents=True, exist_ok=True)

    cacher = Cacher(
        cache_dir,
        invalidate_after_n_minutes=config.invalidate_cache_after_n_minutes,
    )

    if api is None:
        api = CachedOnshapeApi(
            OnshapeClient(),
            cacher,
            max_concurrent_requests=config.max_concurrent_requests,
            post_wait=config.api_post_wait,
        )

    document = api.parse_url(document_url)

    # Runs document pre-checks.
    checked_document = await check_document(document, api, config=config)

    # Converts the checked document to a URDF.
    urdf_info = await save_urdf(checked_document, output_dir, api, config=config)

    return DownloadedDocument(
        check_document=checked_document,
        urdf_info=urdf_info,
    )


async def main(args: Sequence[str] | None = None) -> DownloadedDocument:
    if args is None:
        args = sys.argv[1:]
    config = DownloadConfig.from_cli_args(args)
    configure_logging(level=logging.DEBUG if config.debug else logging.WARN)
    return await download(
        document_url=config.document_url,
        output_dir=config.output_dir,
        config=config,
    )


def sync_main(args: Sequence[str] | None = None) -> None:
    asyncio.run(main(args))


if __name__ == "__main__":
    # python -m onshape.onshape.convert
    sync_main()



"""Defines a pass to flip the orientation of specific joints."""

import argparse
import logging
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Collection

from onshape.formats.common import save_xml

logger = logging.getLogger(__name__)


def _str(i: float) -> str:
    return str(0 if i == 0 else i)


def flip_joints(urdf_path: Path, joint_names: Collection[str]) -> None:
    joint_names = set(joint_names)
    urdf_tree = ET.parse(urdf_path)
    for joint in urdf_tree.iter("joint"):
        if (joint_name := joint.attrib["name"]) in joint_names:
            if (axis := joint.find("axis")) is None or "xyz" not in axis.attrib:
                raise ValueError(f"Joint {joint_name} has invalid axis.")
            new_xyz = [-float(i) for i in axis.attrib["xyz"].strip().split()]
            axis.attrib["xyz"] = " ".join([_str(i) for i in new_xyz])
            if (limit := joint.find("limit")) is not None:
                lower, upper = limit.attrib.get("lower"), limit.attrib.get("upper")
                if lower is not None and upper is not None:
                    limit.attrib["upper"] = _str(-float(lower))
                    limit.attrib["lower"] = _str(-float(upper))
                elif lower is not None or upper is not None:
                    raise ValueError(f"Joint {joint_name} has only one limit.")

            joint_names.remove(joint_name)
    if joint_names:
        options = "\n".join([joint.attrib["name"] for joint in urdf_tree.findall("joint") if "name" in joint.attrib])
        raise ValueError(f"Joints not found: {joint_names} Options are: {options}")
    save_xml(urdf_path, urdf_tree)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF to flip.")
    parser.add_argument("joint_names", nargs="+", help="The names of the joints to flip.")
    args = parser.parse_args()
    flip_joints(args.urdf_path, args.joint_names)



"""Defines geometric utility functions."""

from dataclasses import dataclass

import numpy as np
import stl.mesh
from scipy.spatial import ConvexHull

from onshape.utils.mesh import Mesh


def apply_matrix_(mesh: stl.mesh.Mesh, matrix: np.ndarray) -> stl.mesh.Mesh:
    rotation = matrix[0:3, 0:3]
    translation = matrix[0:3, 3:4].T.tolist()

    def transform(points: np.ndarray) -> np.ndarray:
        return (rotation * np.matrix(points).T).T + translation * len(points)

    mesh.v0 = transform(mesh.v0)
    mesh.v1 = transform(mesh.v1)
    mesh.v2 = transform(mesh.v2)
    mesh.normals = transform(mesh.normals)
    return mesh


def inv_tf(a_to_b_tf: np.matrix) -> np.matrix:
    return np.matrix(np.linalg.inv(a_to_b_tf))


def transform_inertia_tensor(inertia: np.matrix, rotation: np.matrix) -> np.matrix:
    """Transforms the inertia tensor to a new frame.

    Args:
        inertia: The inertia tensor in the original frame, a (3, 3) matrix.
        rotation: The rotation matrix from the original frame to the new frame.

    Returns:
        The inertia tensor in the new frame.
    """
    return rotation @ inertia @ rotation.T


@dataclass
class Dynamics:
    mass: float
    com: np.ndarray
    inertia: np.matrix


def combine_dynamics(*dynamics: Dynamics) -> Dynamics:
    mass: float = 0.0
    com = np.array([0.0] * 3)
    inertia = np.matrix(np.zeros((3, 3)))
    identity = np.matrix(np.eye(3))

    for dynamic in dynamics:
        mass += dynamic.mass
        com += np.array(dynamic.com) * dynamic.mass

    if mass > 0:
        com /= mass

    for dynamic in dynamics:
        r = dynamic.com - com
        p = np.matrix(r)
        inertia = inertia + dynamic.inertia + (np.dot(r, r) * identity - p.T * p) * dynamic.mass

    return Dynamics(mass, com, inertia)


def matrix_to_moments(matrix: np.matrix) -> dict[str, str]:
    return {
        "ixx": str(matrix[0, 0]),
        "ixy": str(matrix[0, 1]),
        "ixz": str(matrix[0, 2]),
        "iyy": str(matrix[1, 1]),
        "iyz": str(matrix[1, 2]),
        "izz": str(matrix[2, 2]),
    }


def moments_to_matrix(inertia_moments: np.ndarray) -> np.matrix:
    """Convert a 6-element array of inertia moments into a 3x3 inertia matrix.

    Args::
        inertia_moments: A 6-element array ``[Ixx, Iyy, Izz, Ixy, Ixz, Iyz]``.

    Returns:
        A (3, 3) inertia matrix.
    """
    ixx, iyy, izz, ixy, ixz, iyz = (
        inertia_moments[0],
        inertia_moments[1],
        inertia_moments[2],
        inertia_moments[3],
        inertia_moments[4],
        inertia_moments[5],
    )

    inertia_matrix = np.array(
        [
            [ixx, ixy, ixz],
            [ixy, iyy, iyz],
            [ixz, iyz, izz],
        ]
    )

    return np.matrix(inertia_matrix)


def get_mesh_convex_hull(mesh: Mesh) -> Mesh:
    hull = ConvexHull(mesh.points)
    return Mesh(points=hull.points, faces=hull.simplices)


def get_center_of_mass(mesh: Mesh) -> tuple[float, float, float]:
    total_volume = 0.0
    center_of_mass = np.zeros(3)
    for triangle in mesh.faces:
        v0, v1, v2 = triangle
        p0, p1, p2 = mesh.points[v0], mesh.points[v1], mesh.points[v2]
        volume = np.dot(p0, np.cross(p1, p2)) / 6.0
        centroid = (p0 + p1 + p2) / 4.0
        total_volume += volume
        center_of_mass += volume * centroid
    center_of_mass /= total_volume
    p0, p1, p2 = center_of_mass
    return (p0, p1, p2)


def scale_mesh(mesh: Mesh, scale: float, about_origin: bool = False) -> Mesh:
    if scale <= 0:
        raise ValueError(f"Scaling {scale} should be greater than 0.")
    com = (0.0, 0.0, 0.0) if about_origin else get_center_of_mass(mesh)
    points = mesh.points - com
    points *= scale
    points += com
    return Mesh(points=points, faces=mesh.faces)


def apply_transform(points: np.ndarray, transform: np.ndarray) -> np.ndarray:
    """Apply a transformation matrix to a set of points.

    Args:
        points: A (n, 3) numpy array of points.
        transform: A (4, 4) transformation matrix.

    Returns:
        A (n, 3) numpy array of transformed points.
    """
    points_homogeneous = np.hstack([points, np.ones((points.shape[0], 1))])
    transformed_points_homogeneous = points_homogeneous @ transform.T
    transformed_points = transformed_points_homogeneous[:, :3]
    transformed_points = np.asarray(transformed_points)
    return transformed_points


def combine_meshes(parent_mesh: Mesh, child_mesh: Mesh, relative_transform: np.ndarray) -> Mesh:
    """Combine parent and child meshes, applying the relative transform to the child mesh points.

    Args:
        parent_mesh: The parent mesh.
        child_mesh: The child mesh.
        relative_transform: The transformation matrix to apply to the child mesh points.

    Returns:
        A new mesh combining the parent and child meshes.
    """
    if not isinstance(relative_transform, np.ndarray) or relative_transform.shape != (4, 4):
        raise ValueError("relative_transform must be a 4x4 numpy array")
    transformed_child_points = apply_transform(child_mesh.points, relative_transform)
    combined_points = np.concatenate([parent_mesh.points, transformed_child_points])
    offset_child_faces = child_mesh.faces + len(parent_mesh.points)
    combined_faces = np.concatenate([parent_mesh.faces, offset_child_faces])
    return Mesh(points=combined_points, faces=combined_faces)



"""Defines a pass to ensure that all inertia matrices are positive definite, above epsilon."""

import argparse
import xml.etree.ElementTree as ET
from pathlib import Path

import numpy as np

DEFAULT_EPSILON = 1e-7  # kg m^2


def fix_inertias(urdf_path: Path, epsilon: float = DEFAULT_EPSILON) -> None:
    urdf_tree = ET.parse(urdf_path)

    # Ensure that all inertia matrices are positive definite by thresholding the diagonal.
    for link in urdf_tree.iter("link"):
        if (inertial := link.find("inertial")) is None:
            continue
        if (inertia := inertial.find("inertia")) is None:
            continue
        try:
            ixx, iyy, izz, ixy, ixz, iyz = [
                float(inertia.attrib[i]) for i in ("ixx", "iyy", "izz", "ixy", "ixz", "iyz")
            ]
        except KeyError:
            continue
        imat = np.array([[ixx, ixy, ixz], [ixy, iyy, iyz], [ixz, iyz, izz]])
        eigvals = np.linalg.eigvals(imat)
        if np.any(eigvals < epsilon):
            raise ValueError(f"Inertia matrix is not positive definite: {eigvals}")


def main() -> None:
    parser = argparse.ArgumentParser(description="Fix inertias in an URDF file.")
    parser.add_argument(
        "urdf_path",
        type=Path,
        help="The path to the URDF file.",
    )
    parser.add_argument(
        "--epsilon",
        type=float,
        default=DEFAULT_EPSILON,
        help="The epsilon value to use for thresholding.",
    )
    args = parser.parse_args()

    fix_inertias(args.urdf_path, epsilon=args.epsilon)


if __name__ == "__main__":
    # python -m onshape.passes.fix_inertias
    main()



"""OnShape client."""

import base64
import datetime
import hashlib
import hmac
import logging
import os
import random
import re
import string
import urllib.parse
import webbrowser
from contextlib import asynccontextmanager
from dataclasses import dataclass
from typing import Any, AsyncIterator, Literal, Mapping, cast

from httpx import AsyncClient, Response

logger = logging.getLogger(__name__)


def escape_url(s: str) -> str:
    return s.replace("/", "%2f").replace("+", "%2b")


ONSHAPE_API_KEY_URL = "https://dev-portal.onshape.com/keys"

DEFAULT_BASE_URL = "https://cad.onshape.com"

Method = Literal["get", "post", "put", "delete"]


WorkspaceType = Literal["w", "v"]


@dataclass
class DocumentInfo:
    document_id: str
    item_kind: WorkspaceType
    item_id: str
    element_id: str

    def get_url(self, base_url: str = DEFAULT_BASE_URL) -> str:
        return f"{base_url}/documents/{self.document_id}/{self.item_kind}/{self.item_id}/e/{self.element_id}"


def get_access_and_secret_keys_interactively() -> tuple[str, str]:
    access_key = os.environ.get("ONSHAPE_ACCESS_KEY")
    secret_key = os.environ.get("ONSHAPE_SECRET_KEY")
    if access_key is not None and secret_key is not None:
        return access_key, secret_key
    print(f"In order to use this tool, you must first set an Onshape API key here: {ONSHAPE_API_KEY_URL}")
    open_web_browser = input("Would you like to open the API key portal? (y/n) ")
    if open_web_browser.lower() in {"y", "yes"}:
        webbrowser.open(ONSHAPE_API_KEY_URL)
    while (access_key_input := input("Enter your access key: ")) == "":
        continue
    while (secret_key_input := input("Enter your secret key: ")) == "":
        continue
    print("For future use, you can set the ONSHAPE_ACCESS_KEY and ONSHAPE_SECRET_KEY environment variables.")
    print("For example:")
    print(f"export ONSHAPE_ACCESS_KEY={access_key_input}")
    print(f"export ONSHAPE_SECRET_KEY={secret_key_input}")
    return access_key_input, secret_key_input


class OnshapeClient:
    def __init__(
        self,
        access_key: str | None = None,
        secret_key: str | None = None,
        base_url: str = DEFAULT_BASE_URL,
        timeout: float = 30.0,
    ) -> None:
        super().__init__()

        if access_key is None or secret_key is None:
            access_key, secret_key = get_access_and_secret_keys_interactively()
        self.access_key = access_key.encode("utf-8")
        self.secret_key = secret_key.encode("utf-8")
        self.base_url = base_url
        self.timeout = timeout

    def parse_url(self, document_url: str) -> DocumentInfo:
        url_match = re.match(rf"{self.base_url}/documents/([\w\d]+)/(w|v)/([\w\d]+)/e/([\w\d]+)", document_url)
        if url_match is None:
            raise ValueError(f"Invalid document URL: {document_url}")
        document_id = url_match.group(1)
        item_kind = cast(WorkspaceType, url_match.group(2))
        item_id = url_match.group(3)
        element_id = url_match.group(4)
        return DocumentInfo(document_id, item_kind, item_id, element_id)

    def _make_nonce(self) -> str:
        """Generate a unique ID for the request, 25 chars in length.

        Returns:
            The unique nonce to use.
        """
        chars = string.digits + string.ascii_letters
        nonce = "".join(random.choice(chars) for i in range(25))
        logger.debug("Created nonce: %s", nonce)
        return nonce

    def _make_auth(
        self,
        method: str,
        date: str,
        nonce: str,
        path: str,
        query: Mapping[str, Any] = {},
        ctype: str = "application/json",
    ) -> str:
        """Create the request signature to authenticate.

        Args:
            method: The HTTP method.
            date: The date header string.
            nonce: The cryptographic nonce.
            path: The URL pathname.
            query: The URL query string in key-value pairs.
            ctype: The HTTP Content-Type.

        Returns:
            The request signature.
        """
        query_str = urllib.parse.urlencode(query)
        hmac_str = f"{method}\n{nonce}\n{date}\n{ctype}\n{path}\n{query_str}\n".lower().encode("utf-8")
        logger.debug("HMAC string: %s", hmac_str)
        signature = base64.b64encode(hmac.new(self.secret_key, hmac_str, digestmod=hashlib.sha256).digest())
        auth = "On " + self.access_key.decode("utf-8") + ":HmacSHA256:" + signature.decode("utf-8")
        logger.debug("Auth header: %s", auth)
        return auth

    def _make_headers(
        self,
        method: str,
        path: str,
        query: Mapping[str, Any] = {},
        headers: Mapping[str, str] = {},
    ) -> dict[str, str]:
        """Creates a headers object to sign the request.

        Args:
            method: The HTTP method.
            path: The URL pathname.
            query: The URL query string in key-value pairs.
            headers: Additional headers to include.

        Returns:
            Dictionary containing all headers
        """
        date = datetime.datetime.now(datetime.UTC).strftime("%a, %d %b %Y %H:%M:%S GMT")
        nonce = self._make_nonce()
        ctype = headers.get("Content-Type", "application/json")
        auth = self._make_auth(method, date, nonce, path, query=query, ctype=ctype)
        req_headers = {
            "Content-Type": "application/json",
            "Date": date,
            "On-Nonce": nonce,
            "Authorization": auth,
            "User-Agent": "Onshape Python Sample App",
            "Accept": "application/json",
        }
        # User-provided headers override defaults.
        for h in headers:
            req_headers[h] = headers[h]
        return req_headers

    @asynccontextmanager
    async def request(
        self,
        method: Method,
        path: str,
        query: Mapping[str, Any] = {},
        headers: Mapping[str, str] = {},
        body: Mapping[str, Any] = {},
        base_url: str | None = None,
    ) -> AsyncIterator[Response]:
        """Issues a request to Onshape.

        Args:
            method: The HTTP method.
            path: The URL pathname.
            query: The URL query string in key-value pairs.
            headers: Additional headers to include.
            body: The request body.
            base_url: The base URL to use.

        Returns:
            The response from Onshape.
        """
        req_headers = self._make_headers(method, path, query, headers)
        if base_url is None:
            base_url = self.base_url
        url = base_url + path + "?" + urllib.parse.urlencode(query)

        async with AsyncClient(
            timeout=self.timeout,
            follow_redirects=False,
        ) as client:
            async with client.stream(
                method,
                url,
                headers=req_headers,
                data=body,
            ) as response:
                if response.status_code == 307:
                    location = urllib.parse.urlparse(response.headers["Location"])
                    querystring = urllib.parse.parse_qs(location.query)
                    logger.debug("Request redirected to: %s", location.geturl())
                    new_query = {}
                    new_base_url = location.scheme + "://" + location.netloc
                    for key in querystring:
                        new_query[key] = querystring[key][0]  # won't work for repeated query params
                    async with self.request(
                        method,
                        location.path,
                        query=new_query,
                        headers=headers,
                        base_url=new_base_url,
                    ) as new_response:
                        yield new_response

                else:
                    if not 200 <= response.status_code <= 206:
                        await response.aread()
                        logger.error("Got response %d for %s, details: %s", response.status_code, path, response.text)

                        if response.status_code == 403:
                            logger.warning(
                                "Check that your access key and secret key are correct, "
                                "and that the document exists and is not private."
                            )

                    else:
                        logger.debug("Got response %d for %s", response.status_code, path)

                    yield response



"""Makes separate collision meshes for each link in the URDF."""

import logging
import shutil
import xml.etree.ElementTree as ET
from pathlib import Path

import trimesh

from onshape.passes.utils import iter_meshes
from onshape.utils.mesh import COLLISION_SUFFIX

logger = logging.getLogger(__name__)


def separate_collision_mesh(visual_mesh_path: Path) -> Path:
    collision_mesh_name = f"{visual_mesh_path.stem}{COLLISION_SUFFIX}{visual_mesh_path.suffix}"
    collision_mesh_path = visual_mesh_path.parent / collision_mesh_name
    if collision_mesh_path.exists():
        logger.warning("Collision mesh %s already exists for %s", collision_mesh_path, visual_mesh_path)
    shutil.copy2(visual_mesh_path, collision_mesh_path)
    return collision_mesh_path


def shrink_collision_mesh(collision_mesh_path: Path, shrink_factor: float) -> None:
    mesh = trimesh.load(collision_mesh_path)
    if not isinstance(mesh, trimesh.Trimesh):
        raise ValueError(f"Loaded mesh {collision_mesh_path} is not a Trimesh")
    mesh.apply_scale(shrink_factor)
    mesh.export(collision_mesh_path)


def shrink_collision_meshes(
    urdf_path: Path,
    shrink_factors: dict[str, float],
) -> None:
    """Shrinks some of the collision meshes in the URDF.

    Args:
        urdf_path: The path to the URDF file.
        shrink_factors: A dictionary mapping link names to shrink factors.
            If the shrink factor is <= 0, the collision mesh will be removed entirely.
    """
    # Make a copy that we can use to keep track of which links have been processed.
    shrink_factors = {k: v for k, v in shrink_factors.items()}

    num_separated = 0
    for link, (_, visual_mesh_path), (col_link, col_mesh_path) in iter_meshes(
        urdf_path,
        save_when_done=True,
    ):
        name = link.attrib["name"]

        if name not in shrink_factors:
            continue

        if col_link is None or col_mesh_path is None:
            raise ValueError(f"No collision mesh found for {name}")

        shrink_factor = shrink_factors.pop(name)

        # If shrink factor is <= 0, remove the collision mesh
        if shrink_factor <= 0:
            logger.info("Removing collision mesh for %s", name)
            col_link_parent = link.find(".//collision")
            if col_link_parent is None:
                raise ValueError(f"No collision link found for {name}")
            link.remove(col_link_parent)
            continue

        # If the collision mesh is the same as the visual mesh, we need to make
        # a new collision mesh.
        if visual_mesh_path is not None and visual_mesh_path == col_mesh_path:
            col_mesh_path = separate_collision_mesh(visual_mesh_path)
            rel_filename = col_mesh_path.relative_to(urdf_path.parent).as_posix()
            col_link.attrib["filename"] = rel_filename
            num_separated += 1

        # After doing this, we can shrink the collision mesh geometry.
        logger.info("Shrinking collision mesh %s by %.2f", col_mesh_path, shrink_factor)
        shrink_collision_mesh(col_mesh_path, shrink_factor)

    if len(shrink_factors) > 0:
        missing_keys = list(shrink_factors.keys())
        urdf_tree = ET.parse(urdf_path)
        options = [link.attrib["name"] for link in urdf_tree.findall(".//link")]
        raise ValueError(f"The following links were not processed: {missing_keys}. Options are {options}")

    if num_separated > 0:
        logger.info(
            "Created %d separate collision meshes for %s",
            num_separated,
            urdf_path.name,
        )



"""Defines common classes and functions to use with schemas."""

import hashlib
from dataclasses import dataclass


@dataclass(frozen=True)
class ElementUid:
    document_id: str
    document_microversion: str
    element_id: str
    part_id: str = ""
    configuration: str = ""

    @property
    def unique_id(self) -> str:
        key = ":".join(
            [
                self.document_id,
                self.document_microversion,
                self.element_id,
                self.part_id,
                self.configuration,
            ],
        )
        return hashlib.sha256(key.encode()).hexdigest()[0:16]



"""Utility functions common to different formats."""

import io
import json
import re
import xml.etree.ElementTree as ET
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Self
from xml.dom import minidom


@dataclass
class JointMetadata:
    actuator_type: str
    id: int
    nn_id: int
    kp: float
    kd: float
    soft_torque_limit: float

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        return cls(**data)

    def to_dict(self) -> dict:
        return {k: v for k, v in asdict(self).items()}

    @classmethod
    def from_json(cls, json_path: Path) -> Self:
        with open(json_path, "r") as f:
            data = json.load(f)
        return cls(**data)


@dataclass
class ActuatorMetadata:
    actuator_type: str
    sysid: str = ""
    max_torque: float = 0.0
    max_velocity: float = 0.0
    armature: float = 0.0
    damping: float = 0.0
    frictionloss: float = 0.0
    vin: float | None = None
    kt: float | None = None
    R: float | None = None
    max_pwm: float | None = None
    error_gain: float | None = None

    def to_dict(self) -> dict:
        return {k: v for k, v in asdict(self).items()}

    @classmethod
    def from_dict(cls, data: dict) -> Self:
        return cls(**data)

    @classmethod
    def from_json(cls, json_path: Path) -> Self:
        with open(json_path, "r") as f:
            data = json.load(f)
        return cls(**data)


def save_xml(path: str | Path | io.StringIO, tree: ET.ElementTree | ET.Element) -> None:
    if isinstance(tree, ET.ElementTree):
        tree = tree.getroot()
    xmlstr = minidom.parseString(ET.tostring(tree)).toprettyxml(indent="  ")
    xmlstr = re.sub(r"\n\s*\n", "\n", xmlstr)
    if isinstance(path, io.StringIO):
        path.write(xmlstr)
    else:
        with open(path, "w") as f:
            f.write(xmlstr)



"""Defines a pass that rotates certain joints."""

import argparse
import logging
import xml.etree.ElementTree as ET
from pathlib import Path
from typing import Collection, Mapping

from scipy.spatial.transform import Rotation as R

from onshape.formats.common import save_xml

logger = logging.getLogger(__name__)


def rotate_joints(urdf_path: Path, joints: Mapping[str, Collection[float | int]]) -> None:
    """Rotates the specified joints by the given RPY values.

    Args:
        urdf_path: The path to the URDF file.
        joints: The joints to rotate, and the RPY values to rotate them by.
    """
    urdf_tree = ET.parse(urdf_path)
    root = urdf_tree.getroot()
    joint_elems = {joint.attrib["name"]: joint for joint in root.findall("joint") if "name" in joint.attrib}

    for joint_name, rpy_vals in joints.items():
        if len(rpy_vals) != 3:
            raise ValueError(f"Expected 3 RPY values for joint {joint_name}, got {len(rpy_vals)}")
        if joint_name not in joint_elems:
            options = "\n".join(joint_elems.keys())
            raise ValueError(f"Joint {joint_name} not found in {urdf_path}. Options are:\n{options}")
        joint = joint_elems[joint_name]
        rpy_tf = R.from_euler("xyz", rpy_vals, degrees=True)
        if (origin := joint.find("origin")) is None:
            raise ValueError(f"Joint {joint_name} has no origin")
        joint_rpy = R.from_euler("xyz", [float(i) for i in origin.attrib["rpy"].split()], degrees=False)
        rpy_tf = rpy_tf * joint_rpy
        rpy_str = " ".join(map(str, rpy_tf.as_euler("xyz", degrees=False)))
        origin.attrib["rpy"] = rpy_str

    save_xml(urdf_path, root)


def main() -> None:
    parser = argparse.ArgumentParser(description="Rotate joints in a URDF file.")
    parser.add_argument("urdf_path", type=Path, help="The path to the URDF file.")
    parser.add_argument("joint", type=str, help="The joint name to rotate.")
    parser.add_argument("rpy", type=float, nargs=3, help="The RPY values to rotate the joint by.")
    args = parser.parse_args()

    r, p, y = args.rpy
    rotate_joints(args.urdf_path, {args.joint: (r, p, y)})


if __name__ == "__main__":
    # python -m onshape.passes.rotate_joints
    main()



"""Defines a cached version of the Onshape API."""

import asyncio
import hashlib
from typing import BinaryIO

from onshape.onshape.api import OnshapeApi
from onshape.onshape.cacher import Cacher
from onshape.onshape.client import DocumentInfo, OnshapeClient, WorkspaceType
from onshape.onshape.schema.assembly import (
    Assembly,
    AssemblyMetadata,
    Part,
    RootAssembly,
    SubAssembly,
)
from onshape.onshape.schema.document import Document
from onshape.onshape.schema.elements import Elements
from onshape.onshape.schema.features import Features
from onshape.onshape.schema.part import PartDynamics, PartMetadata, ThumbnailInfo


class CachedOnshapeApi(OnshapeApi):
    def __init__(
        self,
        client: OnshapeClient,
        cacher: Cacher,
        max_concurrent_requests: int = 1,
        post_wait: float = 0.0,
    ) -> None:
        super().__init__(client, max_concurrent_requests, post_wait)

        self.cacher = cacher

        # Async locks.
        self.master_lock = asyncio.Lock()
        self.stl_locks: dict[str, asyncio.Lock] = {}

    async def get_document(self, did: str) -> Document:
        return await self.cacher(
            f"document_{did}",
            lambda: super(CachedOnshapeApi, self).get_document(did),
            lambda document_str: Document.model_validate_json(document_str),
            lambda document: document.model_dump_json(indent=2),
        )

    async def list_elements(
        self,
        document_id: str,
        workspace_id: str,
        workspace_type: WorkspaceType = "w",
    ) -> Elements:
        return await self.cacher(
            f"elements_{document_id}_{workspace_id}",
            lambda: super(CachedOnshapeApi, self).list_elements(document_id, workspace_id, workspace_type),
            lambda elements_str: Elements.model_validate_json(elements_str),
            lambda elements: elements.model_dump_json(indent=2),
        )

    async def get_first_assembly_id(
        self,
        document_id: str,
        workspace_id: str,
        workspace_type: WorkspaceType = "w",
    ) -> str:
        return await self.cacher(
            f"first_assembly_id_{document_id}_{workspace_id}",
            lambda: super(CachedOnshapeApi, self).get_first_assembly_id(document_id, workspace_id, workspace_type),
            lambda assembly_id: assembly_id,
            lambda assembly_id: assembly_id,
        )

    async def get_assembly(self, document: DocumentInfo, configuration: str = "default") -> Assembly:
        return await self.cacher(
            f"assembly_{document.document_id}_{document.item_id}_{document.element_id}_{configuration}",
            lambda: super(CachedOnshapeApi, self).get_assembly(document, configuration),
            lambda assembly_str: Assembly.model_validate_json(assembly_str),
            lambda assembly: assembly.model_dump_json(indent=2),
        )

    async def get_features(self, asm: RootAssembly | SubAssembly) -> Features:
        return await self.cacher(
            f"features_{asm.key.unique_id}",
            lambda: super(CachedOnshapeApi, self).get_features(asm),
            lambda features_str: Features.model_validate_json(features_str),
            lambda features: features.model_dump_json(indent=2),
        )

    async def get_assembly_metadata(
        self,
        assembly: RootAssembly | SubAssembly,
        configuration: str = "default",
    ) -> AssemblyMetadata:
        return await self.cacher(
            f"assembly_{assembly.key.unique_id}_metadata",
            lambda: super(CachedOnshapeApi, self).get_assembly_metadata(assembly, configuration),
            lambda metadata_str: AssemblyMetadata.model_validate_json(metadata_str),
            lambda metadata: metadata.model_dump_json(indent=2),
        )

    async def get_part_metadata(self, part: Part) -> PartMetadata:
        return await self.cacher(
            f"part_{part.key.unique_id}_metadata",
            lambda: super(CachedOnshapeApi, self).get_part_metadata(part),
            lambda metadata_str: PartMetadata.model_validate_json(metadata_str),
            lambda metadata: metadata.model_dump_json(indent=2),
        )

    async def get_part_mass_properties(self, part: Part) -> PartDynamics:
        return await self.cacher(
            f"part_{part.key.unique_id}_mass_properties",
            lambda: super(CachedOnshapeApi, self).get_part_mass_properties(part),
            lambda props_str: PartDynamics.model_validate_json(props_str),
            lambda props: props.model_dump_json(indent=2),
        )

    async def download_stl(
        self,
        part: Part,
        fp: BinaryIO,
        *,
        units: str = "meter",
        min_facet_width: float | None = None,
        max_facet_width: float | None = None,
        retries: int = 3,
    ) -> None:
        # Each part needs its own lock. We use a master lock to avoid double
        # inserting the same lock.
        async with self.master_lock:
            if part.key.unique_id not in self.stl_locks:
                self.stl_locks[part.key.unique_id] = asyncio.Lock()

        # We need to lock the STL download for the part to avoid downloading
        # the same STL file multiple times.
        async with self.stl_locks[part.key.unique_id]:
            cache_path = self.cacher.cache_dir / f"stl_{part.key.unique_id}.stl"
            if self.cacher._run_fn(cache_path):
                with open(cache_path, "wb") as f:
                    await super(CachedOnshapeApi, self).download_stl(
                        part,
                        f,
                        units=units,
                        min_facet_width=min_facet_width,
                        max_facet_width=max_facet_width,
                        retries=retries,
                    )

        with open(cache_path, "rb") as f:
            fp.write(f.read())

    async def list_thumbnails(self, document: DocumentInfo) -> ThumbnailInfo:
        return await self.cacher(
            f"thumbnails_{document.document_id}_{document.item_id}",
            lambda: super(CachedOnshapeApi, self).list_thumbnails(document),
            lambda thumbnails_str: ThumbnailInfo.model_validate_json(thumbnails_str),
            lambda thumbnails: thumbnails.model_dump_json(indent=2),
        )

    async def download_thumbnail(
        self,
        fp: BinaryIO,
        document: DocumentInfo,
        *,
        width: int = 300,
        height: int = 300,
    ) -> None:
        key = hashlib.md5(f"{document.document_id}_{document.item_id}_{width}_{height}".encode("utf-8")).hexdigest()

        # Each part needs its own lock. We use a master lock to avoid double
        # inserting the same lock.
        async with self.master_lock:
            if key not in self.stl_locks:
                self.stl_locks[key] = asyncio.Lock()

        # We need to lock the STL download for the part to avoid downloading
        # the same STL file multiple times.
        async with self.stl_locks[key]:
            cache_path = self.cacher.cache_dir / f"png_{key}.png"
            if self.cacher._run_fn(cache_path):
                with open(cache_path, "wb") as f:
                    await super(CachedOnshapeApi, self).download_thumbnail(
                        f,
                        document,
                        width=width,
                        height=height,
                    )

        with open(cache_path, "rb") as f:
            fp.write(f.read())



# ruff: noqa: N815
"""Defines the schema for the Document API."""

from pydantic import BaseModel


class DefaultWorkspace(BaseModel):
    id: str
    href: str


class Document(BaseModel):
    defaultWorkspace: DefaultWorkspace



"""Defines the top-level Onshape CLI."""

import argparse
import asyncio
from typing import Literal, Sequence, get_args

from onshape.onshape.download import main as download_main
from onshape.onshape.postprocess import download_and_postprocess_main, postprocess_main
from onshape.scripts.mujoco import mujoco_main
from onshape.scripts.pybullet import pybullet_main

Subcommand = Literal[
    "run",
    "download",
    "postprocess",
    "pybullet",
    "mujoco",
]


async def main(args: Sequence[str] | None = None) -> None:
    parser = argparse.ArgumentParser(description="K-Scale OnShape Library", add_help=False)
    parser.add_argument("subcommand", choices=get_args(Subcommand), help="The subcommand to run")
    parsed_args, remaining_args = parser.parse_known_args(args)
    subcommand: Subcommand = parsed_args.subcommand

    match subcommand:
        case "run":
            await download_and_postprocess_main(remaining_args)
        case "download":
            await download_main(remaining_args)
        case "postprocess":
            await postprocess_main(remaining_args)
        case "pybullet":
            pybullet_main(remaining_args)
        case "mujoco":
            mujoco_main(remaining_args)
        case _:
            raise ValueError(f"Unknown subcommand: {parsed_args.subcommand}")


def sync_main() -> None:
    asyncio.run(main())


if __name__ == "__main__":
    # python -m onshape.scripts.cli
    sync_main()
